/*
 *  @file   FrameQ_ShMem.c
 *
 *  @brief  Implements FrameQ on Shared memory created out of shared region.
 *
 *
 *  ============================================================================
 *
 *  Copyright (c) 2008-2009, Texas Instruments Incorporated
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  
 *  *  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  
 *  *  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  
 *  *  Neither the name of Texas Instruments Incorporated nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  Contact information for paper mail:
 *  Texas Instruments
 *  Post Office Box 655303
 *  Dallas, Texas 75265
 *  Contact information: 
 *  http://www-k.ext.ti.com/sc/technical-support/product-information-centers.htm?
 *  DCMP=TIHomeTracking&HQS=Other+OT+home_d_contact
 *  ============================================================================
 *  
 */



/*-------------------------    XDC  specific includes ----------------------  */
#include <xdc/std.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Startup.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Gate.h>

/*-------------------------  BIOS includes ---------------------------------- */
#include <ti/sysbios/gates/GateMutex.h>

/*-------------------------    Generic includes  ---------------------------- */
#include <string.h>

/*-------------------------    IPC module specific includes ------------------*/
#include <ti/sdo/ipc/Ipc.h>


#include <ti/syslink/inc/ClientNotifyMgr.h>

#ifdef xdc_target__isaCompatible_64P
#include <ti/sysbios/family/c64p/Cache.h>
#else
#include <ti/sysbios/hal/Cache.h>
#endif

#include "package/internal/FrameQ_ShMem.xdc.h"
#include <ti/sdo/utils/_NameServer.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/ipc/_SharedRegion.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/ipc/_ListMP.h>

#include <ti/syslink/ipc/rtos/interfaces/_SysLink.h>
#include <ti/syslink/ipc/rtos/_FrameQBufMgr.h>
#include <ti/syslink/ipc/rtos/_FrameQ.h>
#include <ti/syslink/ipc/rtos/shMem/_FrameQ_ShMem.h>

#define FRAMEQ_NAMESERVERNAME  "FrameQ"

#define ROUND_UP(a, b) (SizeT)((((UInt32) a) + ((b) - 1)) & ~((b) - 1))

extern Ptr gFrameQ_nsHandle;

/*=============================================================================
 * Forward declarations of internal Functions
 *=============================================================================
 */
static inline
Int32
_FrameQ_ShMem_create(FrameQ_ShMem_Object       * obj,
                     const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params * params,
                     Error_Block *eb);

static inline
Int32
_FrameQ_ShMem_open(FrameQ_ShMem_Object       * obj,
                   const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params * params,
                   Error_Block *eb);

static inline Int32
FrameQ_ShMem_readerPutv (FrameQ_ShMem_Object    *obj,
                         FrameQ_ShMem_Frame     framePtr[],
                         UInt32                  filledQueueId[],
                         UInt8                  numFrames);
static inline Int32
FrameQ_ShMem_writerPutv (FrameQ_ShMem_Object    *obj,
                         FrameQ_ShMem_Frame     framePtr[],
                         UInt32                  filledQueueId[],
                         UInt8                  numFrames);
Void FrameQ_Shmem_updateCacheFlags(FrameQ_ShMem_Object *obj,
                                   UInt32               cacheFlags);
/* Called by creator to flush all the remaining frames during delete time */
Int32
FrameQ_ShMem_flushAll (FrameQ_ShMem_Handle handle);
/** ============================================================================
 *  External APIs
 *  ============================================================================
 */

/*
 *  ======== FrameQ_ShMem_getParams ========
 *  API to get the module wide config params.
 */
Int32 FrameQ_ShMem_getConfig( FrameQ_ShMem_Config* cfg)
{
    Int32 status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != cfg), FrameQ_ShMem_A_invalidParams);

    if (NULL != cfg) {
        /* (If setup has not yet been called) */
        if (FrameQ_ShMem_module->refCount == 0) {
            /* Return the default static config values */
            cfg->eventNo        = FrameQ_ShMem_NOTIFY_RESERVED_EVENTNO;
            cfg->maxInstances   = FrameQ_ShMem_MAX_INSTANCES;
        }
        else {
            /* Return the currently configured parameters */
            memcpy (cfg,
                    &FrameQ_ShMem_module->cfg,
                    sizeof (FrameQ_ShMem_Config));
        }

    }
    else {
        status = FrameQ_E_INVALIDARG;
    }

    return (status);
}

/*
 *  ======== FrameQ_setup ========
 *  API to initialize the module dynamically.
 */
Int32
FrameQ_ShMem_setup( FrameQ_ShMem_Config* cfg)
{
    Int32                        status  = FrameQ_S_SUCCESS;
    FrameQ_ShMem_Config    tmpCfg;
    Int32                        key;
    Error_Block                  eb;

    if (cfg == NULL) {
        FrameQ_ShMem_getConfig (&tmpCfg);
        cfg = &tmpCfg;
    }

    /* Check if module is already setup.If not set it up */
    /* Acquire lock */
    key = Gate_enterSystem();

    FrameQ_ShMem_module->refCount++;

    if(FrameQ_ShMem_module->refCount > 1) {
        status = FrameQ_S_ALREADYSETUP;
    }
    else {
        Error_init(&eb);
        /* Check the params required to create the name server instance */
        if (FrameQ_ShMem_module->nameServer == NULL) {
            /* Open  the Name Server instance  created by Frond End FrameQBufMgr
             * module if not  assigned statically.
             */
//            FrameQ_ShMem_module->nameServer =
 //                             NameServer_getHandle(FRAMEQ_NAMESERVERNAME);
            FrameQ_ShMem_module->nameServer = gFrameQ_nsHandle ;
            if(NULL == FrameQ_ShMem_module->nameServer) {
                status = FrameQ_E_OPEN_NAMESERVER;
                /* Decrement the reference count */
                 FrameQ_ShMem_module->refCount--;
            }
        }

    }/* if(FrameQ_ShMem_module->refCount > 1) { */

    /* Release lock */
    Gate_leaveSystem(key);

    return (status);
 }

/*
 *  ======== FrameQ_destroy ========
 *  API to finalize module.
 */
Int32
FrameQ_ShMem_destroy ( )
{
    Int32               status           = FrameQ_S_SUCCESS;
    FrameQ_ShMem_Handle handle     = NULL;
    FrameQ_ShMem_Handle nextHandle = NULL;
    Int32               key;

    /* Acquire lock */
    key = Gate_enterSystem();

    FrameQ_ShMem_module->refCount--;

    if (FrameQ_ShMem_module->refCount == 0) {
        /* Finalize the moduel if it is the last destroy*/

        /* Check if any of the Dynamically created/Opened
         * instances of the module  exist on this processor.
         * Forcefully delete dynamically created objects.
         */
        for ( handle = FrameQ_ShMem_Object_first();
              handle != NULL;
              handle  = nextHandle ) {

            nextHandle = FrameQ_ShMem_Object_next(handle);
            FrameQ_ShMem_delete (&handle);
        }

        FrameQ_ShMem_module->nameServer = NULL;
   }

    /* Release lock */
    Gate_leaveSystem(key);

    return (status);
}



/*
 *  ======== FrameQ_ShMem_Params_init ========
 *  Shim for function to get default initialization param for the FrameQ shMem
 *  instance.
 */
void
FrameQ_ShMem_Params_init(FrameQ_ShMem_Params * params)
{
    ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params lShMemParams;

    ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params_init(&lShMemParams);

    /*Copy contens to shim params */
    memcpy(&(params->commonCreateParams),
           &(lShMemParams.commonCreateParams),
            sizeof (FrameQ_ShMem_Params));
}

/*
 *  ======== FrameQ_sharedMemReq ========
 *  Function to return the amount of shared memory required for
 *  creation of each instance of the ShMem type.
 */
UInt32
FrameQ_ShMem_sharedMemReq (const FrameQ_ShMem_Params *params)
{
    UInt32                 totalSize     = 0;
    ListMP_Params          listMPParams;
    ClientNotifyMgr_Params clientMgrParams;
    GateMP_Params          gateParams;
    SizeT minAlign;
    UInt16 regionId;

    if (NULL != params) {
//        Assert_isTrue(params->numReaders >= FrameQ_MAX_INST_READERS,
//                        FrameQ_ShMem_A_invalidParams);
        if (params->sharedAddr) {
            regionId = SharedRegion_getId(params->sharedAddr);
        }
        else {
            regionId = params->regionId;
        }
        Assert_isTrue(regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_internal);

        minAlign = Memory_getMaxDefaultTypeAlign();
        if (SharedRegion_getCacheLineSize(regionId) > minAlign) {
            minAlign = SharedRegion_getCacheLineSize(regionId);
        }


        totalSize = ROUND_UP(sizeof (FrameQ_NameServerEntry), minAlign);
        /* Attrs size */
        totalSize += ROUND_UP(sizeof (FrameQ_ShMem_Attrs), minAlign);
        /* Writer client */
        totalSize += ROUND_UP(sizeof (FrameQ_ShMem_WriterClient), minAlign);
        /* Reader Clients */
        totalSize += (ROUND_UP(sizeof (FrameQ_ShMem_ReaderClient),minAlign))
                      * params->numReaders;
        /* Reader queues */
        totalSize += (ROUND_UP(sizeof (FrameQ_ShMem_FrameQueue),minAlign))
                     * params->numReaders;

        GateMP_Params_init(&gateParams);
        gateParams.remoteProtect = params->remoteProtect;
        gateParams.localProtect  = params->localProtect;

        if (FrameQ_ShMem_usedefaultgate == FALSE) {
            if(params->gate == NULL) {
                GateMP_Params_init(&gateParams);
                gateParams.regionId = regionId;
                gateParams.remoteProtect =
                                  (GateMP_RemoteProtect) params->remoteProtect;

                gateParams.localProtect  =
                                      (GateMP_LocalProtect)params->localProtect;
                totalSize += ROUND_UP(GateMP_sharedMemReq(&gateParams),
                                       minAlign);/* for self */
                gateParams.localProtect  =
                             (GateMP_LocalProtect)params->localProtect;
                totalSize += ROUND_UP(GateMP_sharedMemReq(&gateParams),
                                       minAlign);/* For ClientNotifyMgr */
            }
        }
        /* Get the default params from the client notify manager */
        ClientNotifyMgr_Params_init(&clientMgrParams);
        clientMgrParams.regionId = regionId;
        /* Number of  notify entries is equal to  number of readers*/
        clientMgrParams.numNotifyEntries = params->numReaders;
        clientMgrParams.numSubNotifyEntries = params->numQueues;
        totalSize += ROUND_UP(ClientNotifyMgr_sharedMemReq (
                                   (ClientNotifyMgr_Params*)&clientMgrParams ),
                              minAlign);

        ListMP_Params_init (&listMPParams);
         /* Set the region id in order to calculate the
          * shared mem required.
          */
        listMPParams.regionId = regionId;

        /* Calculate the shared memory required for the ListMP objects  that are
         * required for all configured readers.
         */
        totalSize += (  ROUND_UP(ListMP_sharedMemReq(&listMPParams),
                                 minAlign)
                      * params->numReaders
                      * params->numQueues);
    }

    return (totalSize);
}

/*
 *  ======== FrameQ_ShMem_getSharedParams ========
 */
static Void FrameQ_ShMem_getSharedParams(FrameQ_ShMem_Params *params,
    const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params *lShMemParams)
{

    Assert_isTrue ((NULL != lShMemParams),FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != params),FrameQ_ShMem_A_invalidParams);

    /*Copy contens to shim params */
    memcpy(&(params->commonCreateParams),
           &(lShMemParams->commonCreateParams),
            sizeof (FrameQ_ShMem_CreateParams));


    params->gate                    = (GateMP_Handle)lShMemParams->gate;
    params->localProtect = (GateMP_LocalProtect)
                                                   lShMemParams->localProtect;
    params->remoteProtect = (GateMP_RemoteProtect)
                                                  lShMemParams->remoteProtect;
    params->sharedAddr              = lShMemParams->sharedAddr;
    params->sharedAddrSize          = lShMemParams->sharedAddrSize;


    params->regionId                = lShMemParams->regionId;
    params->numReaders              = lShMemParams->numReaders;
    params->numQueues               = lShMemParams->numQueues;
    params->cpuAccessFlags          = lShMemParams->cpuAccessFlags ;

    params->frameQBufMgrName        = lShMemParams->frameQBufMgrName;
    params->frameQBufMgrSharedAddr  = lShMemParams->frameQBufMgrSharedAddr ;
}

/*
 *  ======== FrameQ_instance_init ========
 *  API to initialiaze  instance.
 */
Int32
FrameQ_ShMem_Instance_init (FrameQ_ShMem_Object *obj,
                            const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params *params,
                            Error_Block *eb)
{
    Int32                   status      = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != params),FrameQ_ShMem_A_invalidParams);

    if (params->commonCreateParams.openFlag == TRUE) {
        /* Call internal open call */
        status = _FrameQ_ShMem_open(obj, params, eb);
    }
    else {
        /* Call internal open call */
        status = _FrameQ_ShMem_create(obj, params, eb);
    }


    return (status);
}

/*
 *  ======== FrameQ_ShMem_Instance_finalize ========
 *  API to finalize  instance.
 */
void
FrameQ_ShMem_Instance_finalize (FrameQ_ShMem_Object *obj, int instStatus)
{
//    FrameQ_ShMem_ControlStruct *control    = NULL;
    volatile FrameQ_ShMem_Attrs         *pAttrs       = NULL;
    ClientNotifyMgr_Handle               cliMgrHandle = NULL;
    UInt32                               i            = 0;
    volatile FrameQ_ShMem_WriterClient  *writer;
    FrameQ_ShMem_ReaderClient           *reader;
    volatile FrameQ_ShMem_FrameQueue    *frmQueue;
    UInt32                               key;
    UInt32                               k;

    if (NULL != obj) {
        switch (obj->objType) {

        case FrameQ_ShMem_DYNAMIC_OPEN:
            {
                /*
                 *  Update  the shared control info corresponding to this client
                 *  as invalid and Update the other  shared control info.
                 */
                pAttrs = obj->attrs;

                key = GateMP_enter ((GateMP_Handle)obj->gate);

                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_inv((Ptr)pAttrs,
                              sizeof (FrameQ_ShMem_Attrs),
                              Cache_Type_ALL,
                              TRUE);
                }


                if (pAttrs != NULL) {

                   /* Close frame queue lists */
                    for (i = 0; i < pAttrs->numReaders; i++ ) {
                        for (k = 0; k < pAttrs->numQueues; k++ ) {
                            if (obj->frameList[i][k] != NULL) {
                                ListMP_close (
                                    (ListMP_Handle*)&(obj->frameList[i][k]));
                            }
                        }
                    }

                    if (obj->objMode ==  (UInt32)FrameQ_ShMem_MODE_WRITER) {
                        writer = obj->writer;
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_inv((Ptr)writer,
                                      sizeof (FrameQ_ShMem_WriterClient),
                                      Cache_Type_ALL,
                                      TRUE);
                        }
                        /*  Set the writer shared control info as invalid  */
                        writer->isValid    = FALSE;
                        writer->cacheFlags = 0u;
                        writer->procId     = MultiProc_INVALIDID;

                        if (writer->isRegistered == TRUE) {
                            FrameQBufMgr_unregisterNotifier(
                                                      obj->frameQBufMgrHandle);
                        }

                        if (obj->ctrlStructCacheFlag == TRUE) {
                            /* Write back the writer shared control info */
                            Cache_wbInv ((Ptr)writer,
                                         sizeof (FrameQ_ShMem_WriterClient),
                                         Cache_Type_ALL,
                                         TRUE);
                        }
                        obj->objType = 0;

                    }
                    else if (obj->objMode == (UInt32)FrameQ_ShMem_MODE_READER) {
                        /* Point to the correct shared client info */
                        reader =
                            (FrameQ_ShMem_ReaderClient *) obj->virtClientHandle;
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_inv (reader,
                                       sizeof (FrameQ_ShMem_ReaderClient),
                                       Cache_Type_ALL,
                                       TRUE);
                        }
                        if (reader != NULL) {
                            reader->isValid    = FALSE;
                            reader->procId     = MultiProc_INVALIDID;
                            reader->cacheFlags = 0u;

                            frmQueue =  (FrameQ_ShMem_FrameQueue*)
                                             obj->frmQueue[reader->frameQIndex];
                            frmQueue->isAllocated = FALSE;
                            frmQueue->readerIndex = (UInt32)-1;

                            if (reader->isRegistered == TRUE) {
                                ClientNotifyMgr_unregisterClient (
                                                     (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                     reader->notifyId);
                            }

                            /* Do not modify the num elements and list, as  the
                             * queue can be reopened and used.In that  case,
                             * the frames in the correspondig list can be
                             * retrieved and used by the next reader.
                             */
                            pAttrs->numActiveReaders -= 1u;
                        }

                        if (pAttrs->numActiveReaders == 0) {
                            /* TODO: Check do we need to change  primay Queue index
                             * to zero.
                             */

                        }
                        obj->objType = 0;
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_wbInv ((Ptr)pAttrs,
                                         sizeof (FrameQ_ShMem_Attrs),
                                         Cache_Type_ALL,
                                         TRUE);
                            Cache_wbInv ((Ptr)reader,
                                         sizeof (FrameQ_ShMem_ReaderClient),
                                         Cache_Type_ALL,
                                         TRUE);
                            Cache_wbInv ((Ptr)frmQueue,
                                         sizeof (FrameQ_ShMem_FrameQueue),
                                         Cache_Type_ALL,
                                         TRUE);
                        }

                    }
                    else  {
                        /* If mode is other than read or write simply delete the
                         * obejct with out touching the  shared control
                         * structures.This case is possible when setOpenFlags API
                         * has failed.
                         */
                    }
                }

                /* Close client NotifyMgr handle */
                if (obj->clientNotifyMgrHandle != NULL) {
                    cliMgrHandle = (ClientNotifyMgr_Handle) obj->clientNotifyMgrHandle;
                    ClientNotifyMgr_close (&cliMgrHandle);
                }

                if (   (obj->creCliGate == TRUE)
                    && (obj->clientNotifyMgrGate != NULL) ) {
                    /* Close GateMP instance of ClientNotifyMgr.*/
                    GateMP_close ((GateMP_Handle*)&obj->clientNotifyMgrGate);
                }


                for (i = 0; i < pAttrs->numReaders; i++ ) {
                    if (obj->frameList[i] != NULL) {
                        Memory_free(NULL,
                                   (Ptr)obj->frameList[i],
                                    sizeof(UInt32)* pAttrs->numQueues );
                    }
                }
                if (obj->frmQueue != NULL) {
                    Memory_free(NULL,
                                 obj->frmQueue,
                                 (  sizeof(UInt32)
                                  * pAttrs->numReaders
                                  * pAttrs->numQueues));
                }
                if (obj->reader != NULL) {
                    Memory_free(NULL,
                                obj->reader,
                                sizeof(UInt32) * pAttrs->numReaders);
                }

                if (obj->frameList != NULL) {
                    Memory_free(NULL,
                               (Ptr)obj->frameList,
                                sizeof(UInt32)* pAttrs->numReaders);
                }

                GateMP_leave ((GateMP_Handle)obj->gate, key);

                if (obj->name != NULL) {
                    Memory_free(NULL, (Ptr)obj->name, FrameQ_ShMem_MAX_NAMELEN );
                }


                /* Close the instance gate if it is opened internally
                 * in the instance.
                 */
                if ((obj->creGate == TRUE) &&(obj->gate != NULL)) {
                    GateMP_close ((GateMP_Handle*)&obj->gate);
                }
                /* Close the FrameQBufMgr instance opened by this client.*/
                if (obj->frameQBufMgrHandle != NULL) {
                    FrameQBufMgr_close(&obj->frameQBufMgrHandle);
                }

            }
            break;

        case FrameQ_ShMem_DYNAMIC_CREATE:
        case FrameQ_ShMem_DYNAMIC_CREATE_USEDREGION:
            {

                pAttrs =  obj->attrs;

                key = GateMP_enter ((GateMP_Handle)obj->gate);

                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_inv((Ptr)pAttrs,
                              sizeof (FrameQ_ShMem_Attrs),
                              Cache_Type_ALL,
                              TRUE);
                }
                pAttrs->status = !FrameQ_ShMem_CREATED;

                /* Flush all the remaining frames */
                FrameQ_ShMem_flushAll(obj);

               /* Close frame queue lists */
                for (i = 0; i < pAttrs->numReaders; i++ ) {
                    for (k = 0; k < pAttrs->numQueues; k++ ) {
                        if (obj->frameList[i][k] != NULL) {
                            ListMP_delete (
                                  (ListMP_Handle*)&(obj->frameList[i][k]));
                        }
                    }
                }
                for (i = 0; i < pAttrs->numReaders; i++ ) {
                    if (obj->frameList[i] != NULL) {
                        Memory_free(NULL,
                                   (Ptr)obj->frameList[i],
                                    sizeof(UInt32)*
                                                pAttrs->numQueues );
                    }
                }

                if (obj->frameList != NULL) {
                    Memory_free(NULL,
                               (Ptr)obj->frameList,
                                sizeof(UInt32)* pAttrs->numReaders);
                }
                if (obj->frmQueue != NULL) {
                    Memory_free(NULL,
                                 obj->frmQueue,
                                 (  sizeof(UInt32)
                                  * pAttrs->numReaders
                                  * pAttrs->numQueues));
                }
                if (obj->reader != NULL) {
                    Memory_free(NULL,
                                obj->reader,
                                sizeof(UInt32) * pAttrs->numReaders);
                }

                memset ((Ptr)pAttrs, 0, sizeof(FrameQ_ShMem_Attrs));
                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_wbInv ((Ptr)pAttrs,
                                 sizeof (FrameQ_ShMem_Attrs),
                                 Cache_Type_ALL,
                                 TRUE);
                }


                if (obj->clientNotifyMgrHandle != NULL) {
                    cliMgrHandle = (ClientNotifyMgr_Handle) obj->clientNotifyMgrHandle;
                    ClientNotifyMgr_delete (&cliMgrHandle);
                }

                /* Delete GateMP instance of ClientNotifyMgr.*/
                if (   (obj->clientNotifyMgrGate != NULL)
                    && (obj->creCliGate == TRUE)) {
                    GateMP_delete ((GateMP_Handle*)&obj->clientNotifyMgrGate);
                }

                /* Close the FrameQBufMgr instance opened by this client.*/
                if (obj->frameQBufMgrHandle != NULL) {
                    FrameQBufMgr_close(&obj->frameQBufMgrHandle);
                }

                /* Remove entry from NameServer */
                if (   (NULL != FrameQ_ShMem_module->nameServer)
                    && (obj->nsKey != 0)) {
                    NameServer_removeEntry((NameServer_Handle)FrameQ_ShMem_module->nameServer,
                                           obj->nsKey);
                }

                GateMP_leave ((GateMP_Handle)obj->gate, key);

                /* Free the allocated memory for the instance control structure
                 */
                if (obj->objType == FrameQ_ShMem_DYNAMIC_CREATE_USEDREGION) {
                    Memory_free(SharedRegion_getHeap(obj->regionId),
                                obj->pnameserverEntry,
                                obj->allocSize);
                }

                /* Delete instance gate if it is opened internally
                 * in the instance.
                 */
                if (   (obj->creGate == TRUE)
                    && (obj->gate != NULL)) {
                    GateMP_delete ((GateMP_Handle*)&obj->gate);
                    /* Free the allocated memory for the instance control
                     * structure.
                     */
                    Memory_free(SharedRegion_getHeap(obj->regionId),
                                obj->instGateSharedAddr,
                                obj->gateMPAllocSize);
                }

                if (obj->name != NULL) {
                    Memory_free(NULL, (Ptr)obj->name, FrameQ_ShMem_MAX_NAMELEN );
                }
            }

            break;

        case FrameQ_ShMem_STATIC_OPEN :
            {
                /* Statically opened instances can not be finalized */
            }
            break;

        case FrameQ_ShMem_STATIC_CREATE:
            {
                /* Statically created instances can not be finalized */
            }
            break;

        default:
            {
                /* Object does  not have proper object type defined*/
            }
            break;
        }
    }
}

/*
 *  ======== FrameQ_ShMem_open ========
 *  API to open the FrameQ instance.
 */
Int32
FrameQ_ShMem_open (ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_OpenParams *openParams,
                   FrameQ_ShMem_Handle *handlePtr)

{
    FrameQ_ShMem_Handle  handle    = NULL;
    Int32                status    = FrameQ_S_SUCCESS;
    Int32                tmpStatus = FrameQ_S_SUCCESS;
    ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params  params ;
    Error_Block          eb;

    Error_init(&eb);
    ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params_init (&params) ;

    /* This is open call . So Updating the params accordingly */
    params.commonCreateParams.openFlag       = TRUE ;
    /* Copy name */
    Assert_isTrue((   strlen((String)openParams->commonOpenParams.name)
                    < FrameQ_ShMem_MAX_NAMELEN),
                  FrameQ_ShMem_A_invalidParams);

    /* Name is string pointer in CreateParams*/
    params.commonCreateParams.name = (String)
                                            openParams->commonOpenParams.name;
    /* Not passing shared addr */
    params.sharedAddr         = openParams->commonOpenParams.sharedAddr;
    params.cpuAccessFlags     = openParams->commonOpenParams.cpuAccessFlags;

    handle = FrameQ_ShMem_create (&params, &eb);

    if (handle == NULL) {
        status = FrameQ_E_FAIL;
        *handlePtr = NULL ;
    }
    else {
        /* Prepare this handle for the specified openParams */
        tmpStatus = FrameQ_ShMem_setOpenFlags (handle,
               (ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_OpenParams*)openParams);
        if ( tmpStatus < 0) {
            status = tmpStatus;
            /*Delete the created FrameQ */
            FrameQ_ShMem_delete (&handle);
            *handlePtr = NULL;
        }
        else {
            *handlePtr = handle;
        }
    }
    return (status);
}

/*
 *  ======== FrameQ_ShMem_close ========
 *  API to close the FrameQ.
 */
Int32
FrameQ_ShMem_close (FrameQ_ShMem_Handle *phandle)
{
    Int32 status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != phandle) ,
                    FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != *phandle) ,
                    FrameQ_ShMem_A_invalidParams);
    /* Call  delete here.
     * Delete forcefully delets the client  .
     */
    FrameQ_ShMem_delete (phandle);

    return (status);
}

/*
 *  ======== FrameQ_ShMem_alloc ========
 * Function to allocate a  frame.
 * It allocates frame from the free Queue 0 of FrameQBufMgr.
 */

Int32
FrameQ_ShMem_alloc (FrameQ_ShMem_Object  *obj,
                    FrameQ_ShMem_Frame   *framePtr)
{
    Int32   status  = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr),
                   FrameQ_ShMem_A_invalidParams);

    if (obj->objMode != FrameQ_ShMem_MODE_WRITER) {
        /* If Caller is other than writer, return error. */
        status    = FrameQ_E_ACCESSDENIED;
        *framePtr = NULL;
    }
    else {
        /* Get the frame from free queue 0 of FrameQ bufMgr. */
        status = FrameQBufMgr_alloc (obj->frameQBufMgrHandle,
                                     (FrameQBufMgr_Frame *)framePtr);
    }
    return (status);
}

/*
 *  ======== FrameQ_ShMem_allocv ========
 * Function to allocate multiple frames from the free frame queues given
 * (free frame queues are configured in FrameQbufMgr).
 * After API returns,numframes will denotes  the number of successfully
 * allocated frames.
 */
Int32
FrameQ_ShMem_allocv (FrameQ_ShMem_Object *obj,
                     FrameQ_ShMem_Frame   framePtr[],
                     UInt32               freeQId[],
                     UInt8               *numFrames)
{
    Int32 status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != freeQId),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFrames),
                   FrameQ_ShMem_A_invalidParams);

    if (obj->frameQBufMgrHandle == NULL) {
        status = FrameQ_E_FAIL;
    }
    else if (obj->objMode != FrameQ_ShMem_MODE_WRITER) {
        /* If Caller is other than writer, return error. */
        status = FrameQ_E_ACCESSDENIED;

    }
     else {
         /* Allocate  frames from the plugged in FrameQBufMgr instance.
          * Base frame headers need not be populated  here as FrameQBufMgr will
          * return the frames  having updated headers.
          *
          *  If Alloc fails FrameQbufMgr takes care of enablling notifications
          *  for the writer clients.
          */
         status = FrameQBufMgr_allocv (obj->frameQBufMgrHandle,
                                       (FrameQBufMgr_Frame *)framePtr,
                                       freeQId,
                                       numFrames);
    }

    return (status);
}

/*
 *  ======== FrameQ_ShMem_free ========
 * Function to free single frame
 */
Int32
FrameQ_ShMem_free (FrameQ_ShMem_Object *obj, FrameQ_ShMem_Frame  frame)
{
    Int32   status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue (( (obj->objMode == FrameQ_ShMem_MODE_WRITER)
                    ||(obj->objMode == FrameQ_ShMem_MODE_READER)),
                   FrameQ_ShMem_A_invalidParams);

     status = FrameQBufMgr_free(obj->frameQBufMgrHandle,
                                   (FrameQBufMgr_Frame)frame);

    return (status);
}

/*
 *  ======== FrameQ_freev ========
 * Function to free  multiple frames. API assumes that all frames are allocated
 * from the same frameQbufMgr.i.e  These frames are allocated by FrameQ writers
 * whose plugged in FrameQBufMgr is same.
 */
Int32
FrameQ_ShMem_freev (FrameQ_ShMem_Object *obj,
                    FrameQ_ShMem_Frame   framePtr[],
                    UInt32               numFrames)
{
    Int32               status = FrameQ_S_SUCCESS;
    FrameQBufMgr_Handle bufMgrHandle;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((0 != numFrames),
                   FrameQ_ShMem_A_invalidParams);

    /* Currently assuming all frames from the same buffer manager allocater*/
    if (    (obj->objMode != FrameQ_ShMem_MODE_WRITER)
         && (obj->objMode != FrameQ_ShMem_MODE_READER)) {
        /* If Caller is other than writer or reader, return error. */
        status = FrameQ_E_ACCESSDENIED;
    }
    else {
        bufMgrHandle = FrameQBufMgr_getHandle (framePtr[0]->frmAllocaterId);
        Assert_isTrue ((NULL != bufMgrHandle),
                       FrameQ_ShMem_A_invalidFrameQBufMgrId);
        if (bufMgrHandle == NULL) {
            status = FrameQ_E_FAIL;
        }
        else {
            status = FrameQBufMgr_freev(bufMgrHandle,
                                        (FrameQBufMgr_Frame *)framePtr,
                                        numFrames);
        }
    }

    return (status);
}

/*
 * ======== FrameQ_put ========
 * Function to  insert frame in to FrameQ. It inserts frame in to filled
 * frames queue 0 of FrameQ reader client.
 */
Int32
FrameQ_ShMem_put  (FrameQ_ShMem_Object *obj, FrameQ_ShMem_Frame  frame)
{
    Int32                  status = FrameQ_S_SUCCESS;
    UInt32                 i            = 0;
    UInt8                  filledQueueNo = 0;
    Bool                    tonotify = FALSE;
//    volatile FrameQ_ShMem_Attrs         *pAttrs       = NULL;
    FrameQ_ShMem_Frame      pframe[FrameQ_MAX_INST_READERS] = {NULL};

    volatile FrameQ_ShMem_ReaderClient  *reader;
    volatile FrameQ_ShMem_FrameQueue    *queue;
    UInt32                 counter;
    FrameQ_ShMem_Frame     temp;
    UInt32                 key;
    Int32                  tmpStatus;
    FrameQBufMgr_Handle    bufMgrHandle;
    Ptr                    portablePtr;
    UInt32                 k;

    Assert_isTrue ((NULL != obj), FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);

    Assert_isTrue (((obj->objMode == FrameQ_ShMem_MODE_READER)
                    || (obj->objMode == FrameQ_ShMem_MODE_WRITER)),
                    FrameQ_ShMem_A_invalidParams);

    filledQueueNo  = 0;
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)obj->attrs,
                   sizeof (FrameQ_ShMem_Attrs),
                   Cache_Type_ALL,
                   TRUE);
    }
    if (obj->objMode == FrameQ_ShMem_MODE_READER) {
        reader = obj->virtClientHandle;
        queue = (FrameQ_ShMem_FrameQueue *)
                                         obj->frmQueue[reader->frameQIndex];
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)queue,
                       sizeof (FrameQ_ShMem_FrameQueue),
                       Cache_Type_ALL,
                       TRUE);
        }
        if (queue->isAllocated != TRUE) {
            status = FrameQ_E_FAIL;
        }
        else {
            /* If client happens to be  reader, then just put the frame at
             * the head of the corresponding reader's queue
             */
            /* Convert all address to sharead region address in
             * frame header.i.e frame buffer pointers
             */
            bufMgrHandle = obj->frameQBufMgrHandle;
            Assert_isTrue ((NULL != bufMgrHandle),
                           FrameQ_ShMem_A_invalidFrameQBufMgrId);

            for (k = 0; k < frame->numFrameBuffers; k++) {
                if (   (obj->frmBufCacheFlag[k] == TRUE)
                    && (obj->bufCpuAccessFlag[k] == TRUE)) {
                status = FrameQBufMgr_writeBackFrameBuf(
                                bufMgrHandle,
                               (Ptr)
       (frame->frameBufInfo[k].bufPtr + frame->frameBufInfo[k].startOffset),
                               frame->frameBufInfo[k].validSize,
                               k);
                }
                /* Address translation to portable pointers */
                status =  FrameQBufMgr_translateAddr (
                                    bufMgrHandle,
                                    &portablePtr,
                                    FrameQBufMgr_AddrType_Portable,
                                    (Ptr)frame->frameBufInfo[k].bufPtr,
                                    FrameQBufMgr_AddrType_Virtual,
                                    FrameQBufMgr_BUF_FRAMEBUF);

                Assert_isTrue ( (NULL != portablePtr),
                                 FrameQ_ShMem_A_nullPointer);
                frame->frameBufInfo[k].bufPtr = (UInt32)portablePtr;
            }

            if (obj->frmHdrBufCacheFlag == TRUE) {
                status = FrameQBufMgr_writeBackHeaderBuf(bufMgrHandle,
                                                         frame);
            }
            /* Put in to filled queue 0 of the reader */
            ListMP_putHead ((ListMP_Handle)obj->frameList[reader->frameQIndex][filledQueueNo] ,
                           (ListMP_Elem *)frame);
            queue->numFrames[filledQueueNo]++;

            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_wbInv ((Ptr)queue,
                             sizeof (FrameQ_ShMem_FrameQueue),
                             Cache_Type_ALL,
                             TRUE);
            }
        }


    }
    else {
        bufMgrHandle = obj->frameQBufMgrHandle;
        Assert_isTrue ((NULL != bufMgrHandle),
                       FrameQ_ShMem_A_invalidFrameQBufMgrId);

        if (obj->attrs->numActiveReaders > 1) {
            /*More than one reader exist. Need to duplicate frame to insert
             *frames in to queues other than primary queue.
             */
             /* Allocate frames  by duplicating. Need
              * to allocate as many as there are number of active readers
              * except primary queue reader.
              */

            tmpStatus = FrameQBufMgr_dup (bufMgrHandle,
                                          (FrameQBufMgr_Frame)frame,
                                          (FrameQBufMgr_Frame *)pframe,
                                          obj->attrs->numActiveReaders);
            if (tmpStatus < 0) {
                status = tmpStatus;
            }
        }
        if (status >= 0) {
            /* Write back the original frame */

            /* Translate the frame buffer address to portable addreess */
            for (k = 0; k < frame->numFrameBuffers; k++) {
                if (   (obj->frmBufCacheFlag[k] == TRUE)
                    && (obj->bufCpuAccessFlag[k] == TRUE)) {
                    status = FrameQBufMgr_writeBackFrameBuf(
                               bufMgrHandle,
                               (Ptr)
       (frame->frameBufInfo[k].bufPtr + frame->frameBufInfo[k].startOffset),
                               frame->frameBufInfo[k].validSize,
                               k);
                }
                /* Address translation to portable pointers */
                status =  FrameQBufMgr_translateAddr (
                                    bufMgrHandle,
                                    &portablePtr,
                                    FrameQBufMgr_AddrType_Portable,
                                    (Ptr)frame->frameBufInfo[k].bufPtr,
                                    FrameQBufMgr_AddrType_Virtual,
                                    FrameQBufMgr_BUF_FRAMEBUF);

                Assert_isTrue ( (NULL != portablePtr),
                                 FrameQ_ShMem_A_nullPointer);
                frame->frameBufInfo[k].bufPtr = (UInt32)portablePtr;
            }

            if (obj->frmHdrBufCacheFlag == TRUE) {
                status = FrameQBufMgr_writeBackHeaderBuf(bufMgrHandle,
                                                         frame);
            }

            /* Copy the contents of original frame header contents to other
             * duped frame headers.
             */
            if (obj->attrs->numActiveReaders > 1) {
                for (i = 0; i < obj->attrs->numActiveReaders -1; i++) {
                    memcpy (pframe[i], frame, frame->headerSize);
                }
            }
        }

        if (obj->attrs->numActiveReaders > 1) {
             counter = 0;

             for (  i = 0; i < obj->attrs->numReaders; i++) {
                if( i != obj->attrs->primaryQueueId) {
                    queue = obj->frmQueue[i];
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_inv((Ptr)queue,
                                  sizeof (FrameQ_ShMem_FrameQueue),
                                  Cache_Type_ALL,
                                  TRUE);
                    }

                    temp = (FrameQ_ShMem_Frame)pframe[i];
                }
                else {
                    /* Insert the original frame in to primary queue
                     * irrespective of numActive readers
                     */
                    queue = obj->frmQueue[obj->attrs->primaryQueueId];
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_inv((Ptr)queue,
                                  sizeof (FrameQ_ShMem_FrameQueue),
                                  Cache_Type_ALL,
                                  TRUE);
                    }
                    temp = frame;

                }
                if (   (queue->isAllocated == TRUE)
                    || (i == obj->attrs->primaryQueueId)) {
                    /* Allow insertion in to primary queue evene if it
                     * is not allocated to any reader.
                     */
                    counter++;
                    /* Increment the frame count of filled queue 0 of
                     * reader.
                     */
                    queue->numFrames[filledQueueNo]++;
                    /* Since frame headers are also from share region,
                     * Internal ListMP operation on list element will not be
                     * a problem(ListMP calls Shared region APIs on the list
                     * element).
                     */
                    ListMP_putTail ((ListMP_Handle)obj->frameList[i][filledQueueNo] ,
                                    (ListMP_Elem *)temp);
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_wbInv((Ptr)queue,
                                    sizeof (FrameQ_ShMem_FrameQueue),
                                    Cache_Type_ALL,
                                    TRUE);
                    }
                    /* Send Notification to the Reader if it
                     * is registered for notification
                     */
                    if (obj->attrs->numActiveReaders > 0) {
                        reader = obj->reader[queue->readerIndex];
                        if (obj->ctrlStructCacheFlag == TRUE) {
                             Cache_inv((Ptr)reader,
                                       sizeof (FrameQ_ShMem_ReaderClient),
                                       Cache_Type_ALL,
                                       TRUE);
                        }
                        tonotify = TRUE;
                    }
                    if ((tonotify == TRUE)&&(reader->isRegistered == TRUE)) {
                        /* Release lock */
                        GateMP_leave ((GateMP_Handle)obj->gate, key);
                        status =
                            ClientNotifyMgr_sendNotification (
                                             (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                             reader->notifyId,
                                             filledQueueNo,
                                             queue->numFrames[filledQueueNo],
                                            (UInt32)-1);
                        key = GateMP_enter ((GateMP_Handle)obj->gate);

                    }
                }

                if (counter == obj->attrs->numActiveReaders) {
                    break;
                }
             }
        }
        else {
           /* Insert the original frame in to primary queue
            */
            queue = obj->frmQueue[obj->attrs->primaryQueueId];
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv((Ptr)queue,
                          sizeof (FrameQ_ShMem_FrameQueue),
                          Cache_Type_ALL,
                          TRUE);
            }
            temp = frame;
            queue->numFrames[filledQueueNo]++;

            /* Check if queue is allocated  to reader then only do invalidation
             * of reader to send notifications. There may be a case
             * where reader is not opened yet.
             */
            if (queue->isAllocated == TRUE) {
                reader = (FrameQ_ShMem_ReaderClient*)
                                                obj->reader[queue->readerIndex];
                if (obj->ctrlStructCacheFlag == TRUE) {
                     Cache_inv((Ptr)reader,
                               sizeof (FrameQ_ShMem_ReaderClient),
                               Cache_Type_ALL,
                               TRUE);
                }
                tonotify = TRUE;
            }

            ListMP_putTail ((ListMP_Handle)obj->frameList[obj->attrs->primaryQueueId][filledQueueNo] ,
                           (ListMP_Elem *)temp);
            if (obj->ctrlStructCacheFlag == TRUE) {
               Cache_wbInv((Ptr)queue,
                           sizeof (FrameQ_ShMem_FrameQueue),
                           Cache_Type_ALL,
                           TRUE);
            }

            /* Check and send notification to  the corresponding reader
            * only if notification is registered.
            */
            if ((tonotify == TRUE) && (reader->isRegistered == TRUE)) {
                /* Release lock */
                GateMP_leave ((GateMP_Handle)obj->gate, key);
                status =
                    ClientNotifyMgr_sendNotification (
                                             (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                             reader->notifyId,
                                             filledQueueNo,
                                             queue->numFrames[filledQueueNo],
                                             (UInt32)-1);
                return(status);

            }
        }
    }
    /* Release lock */
    GateMP_leave ((GateMP_Handle)obj->gate, key);
    return (status);
}

/*
 * ======== FrameQ_putv ========
 * Function to insert multiple frames  in to queues of a
 * FrameQ reader client. If multiple readers exist, It dups the frames and
 * insert in to other reader queues.
 */
Int32
FrameQ_ShMem_putv (FrameQ_ShMem_Object    *obj,
                   FrameQ_ShMem_Frame     framePtr[],
                   UInt32                  filledQueueId[],
                   UInt8                  numFrames)
{
    Int32                  status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != filledQueueId),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFrames),
                   FrameQ_ShMem_A_invalidParams);

    if (   (obj->objMode != FrameQ_ShMem_MODE_READER)
        && (obj->objMode != FrameQ_ShMem_MODE_WRITER)) {
        /* The provided handle does not have permission to do  put */
        status = FrameQ_E_ACCESSDENIED;
    }
    else if (obj->objMode == FrameQ_ShMem_MODE_READER){
        status = FrameQ_ShMem_readerPutv(obj,
                                         framePtr,
                                         filledQueueId,
                                         numFrames);
    }
    else  {
        status = FrameQ_ShMem_writerPutv(obj,
                                         framePtr,
                                         filledQueueId,
                                         numFrames);
    }
    return (status);
}

/*
 *  ======== FrameQ_ShMem_get ========
 *  API to get frame  from FrameQ.
 *
 *  Allow only readers to call  this API.
 */
Int32
FrameQ_ShMem_get (FrameQ_ShMem_Object *obj, FrameQ_ShMem_Frame *framePtr)
{
    Int32                        status  = FrameQ_S_SUCCESS;
    volatile FrameQ_ShMem_Attrs          *pAttrs;
    FrameQ_ShMem_ReaderClient   *reader;
    volatile FrameQ_ShMem_FrameQueue     *queue;
    FrameQ_ShMem_Frame           frame;
    Ptr                          bufPtr;
    Int32                        key;
    UInt32                       i;
    UInt8                        filledQueueNo;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams );

    Assert_isTrue ((NULL != framePtr),
                   FrameQ_ShMem_A_invalidParams );

    if (obj->objMode != FrameQ_ShMem_MODE_READER) {
        status = FrameQ_E_ACCESSDENIED;
    }
    else {
        pAttrs =  obj->attrs;
        /* Get simply retrieves frame from the filled queue 0 of the reader */
        filledQueueNo = 0;
        reader = (FrameQ_ShMem_ReaderClient *) obj->virtClientHandle;

        /* Acquire  lock */
        key = GateMP_enter ((GateMP_Handle)obj->gate);
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv (reader,
                       sizeof (FrameQ_ShMem_ReaderClient),
                       Cache_Type_ALL,
                       TRUE);
        }

        if(reader->isValid != TRUE) {
            status = FrameQ_E_ACCESSDENIED;
        }
        else {
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv ((Ptr)pAttrs,
                           sizeof (FrameQ_ShMem_Attrs),
                           Cache_Type_ALL,
                           TRUE);
            }
            queue = obj->frmQueue[reader->frameQIndex];
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv ((Ptr)queue,
                           sizeof (FrameQ_ShMem_FrameQueue),
                           Cache_Type_ALL,
                           TRUE);
            }

            if (queue->numFrames[filledQueueNo] == 0 ) {
                status = FrameQ_E_EMPTY;
            }
            else {
                frame = (FrameQ_ShMem_Frame)ListMP_getHead ((ListMP_Handle)
                            obj->frameList[reader->frameQIndex][filledQueueNo]);

                Assert_isTrue ((NULL != frame),FrameQ_ShMem_A_nullPointer);

                if (frame != NULL) {
                    queue->numFrames[filledQueueNo] -= 1;
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_wbInv ((Ptr)queue,
                                     sizeof (FrameQ_ShMem_FrameQueue),
                                     Cache_Type_ALL,
                                     TRUE);
                    }

                    /* Invalidate  frame header.
                     * Conevert portable frame pointers to virtual pointers.
                     * Then do invalidate frame buffers.
                     */
                    if (obj->frmHdrBufCacheFlag == TRUE){
                        status = FrameQBufMgr_invalidateHeaderBuf(
                                                    obj->frameQBufMgrHandle,
                                                    frame);
                    }

                    /* Convert  portable address of framebuffers to  virtual
                     * address in the retrieved frame.
                     */
                     for ( i = 0; i < frame->numFrameBuffers; i++) {
                        /* Address translation to virtual pointers */
                        status =  FrameQBufMgr_translateAddr (
                                            obj->frameQBufMgrHandle,
                                            &bufPtr,
                                            FrameQBufMgr_AddrType_Virtual,
                                            (Ptr)frame->frameBufInfo[i].bufPtr,
                                            FrameQBufMgr_AddrType_Portable,
                                            FrameQBufMgr_BUF_FRAMEBUF);

                        Assert_isTrue ( (NULL != bufPtr),
                                         FrameQ_ShMem_A_nullPointer);
                        frame->frameBufInfo[i].bufPtr = (UInt32)bufPtr;

                        if (   (obj->frmBufCacheFlag[i] == TRUE)
                           && (obj->bufCpuAccessFlag[i] == TRUE)) {
                            /* Invalidate  frame buffers pointed to by this
                             * frame Using optimum cache coherence operations
                             */
                            status = FrameQBufMgr_invalidateFrameBuf(
                                            obj->frameQBufMgrHandle,
                                           (Ptr)(frame->frameBufInfo[i].bufPtr +
                                            frame->frameBufInfo[i].startOffset),
                                            frame->frameBufInfo[i].validSize,
                                            i);
                        }
                        *framePtr = frame;
                    }

                }
                else {
                    status = FrameQ_E_FAIL;
                }
            }
        }

        /* Release lock */
        GateMP_leave ((GateMP_Handle)obj->gate , key);

        if (status < 0) {
            *framePtr = NULL;
            if (    (reader->isRegistered == TRUE)
                 && (reader->notifyId != (UInt32)-1)) {
                ClientNotifyMgr_enableNotification ( (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                     reader->notifyId,
                                                     filledQueueNo);
            }
        }
    }
    return (status);
}

/*
 * ======== FrameQ_ShMem_getv ========
 * Function to retrieve frames from  queues of a FrameQ Reader client.
 * It returns frame if it is available in the queue.
 */
Int32
FrameQ_ShMem_getv (FrameQ_ShMem_Object   *obj,
                   FrameQ_ShMem_Frame     pframe[],
                   UInt32                 filledQueueId[],
                   UInt8                 *numFrames)
{
    Int32                        status     = FrameQ_S_SUCCESS;
    UInt32                       lNumFrames = 0;
    volatile FrameQ_ShMem_Attrs *pAttrs;
    FrameQ_ShMem_ReaderClient   *reader;
    volatile FrameQ_ShMem_FrameQueue     *queue;
    FrameQ_ShMem_Frame           frame;
    Ptr                          bufPtr;
    Int32                        key;
    UInt32                       i;
    UInt32                       j;
    UInt8                        filledQueueNo;
    Int32                        tmpStatus;
    /* TO find out the duplicate freeQ ids and frames requested in that free Q*/
    UInt32                       pfilledQIds[FrameQ_MAX_FRAMESINVAPI]= {0};
    UInt32                       pNumFrames[FrameQ_MAX_FRAMESINVAPI]= {0};
    UInt32                       count = 0;
    UInt32                       pNotifySubIds[FrameQ_MAX_FRAMESINVAPI] = {0};
    UInt32                       numSubEntries = 0;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != pframe),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != filledQueueId),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFrames),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((FrameQ_ShMem_MODE_READER == obj->objMode),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((FrameQ_MAX_FRAMESINVAPI >= *numFrames),
                   FrameQ_ShMem_A_invalidParams);

    /* Find out the duplicate entries in freeQids specified */
    for (i = 0 ; i < *numFrames ; i++) {
        if (pfilledQIds [i] != (Uint32) -1) {
            pfilledQIds [count] = filledQueueId[i];
            pNumFrames [count] += 1u;

            for (j = i + 1 ; j < *numFrames ; j++) {
                if (pfilledQIds [count] ==  filledQueueId[j]) {
                    pNumFrames [count] += 1;
                    pfilledQIds [j] = (Uint32) -1 ;
                }
            }
            count++ ;
        }
    }

    pAttrs =  obj->attrs;
    reader = (FrameQ_ShMem_ReaderClient *) obj->virtClientHandle;
    /* Acquire  lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv (reader,
                   sizeof (FrameQ_ShMem_ReaderClient),
                   Cache_Type_ALL,
                   TRUE);
    }

    if(reader->isValid != TRUE) {
        status = FrameQ_E_ACCESSDENIED;
    }
    else {
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)pAttrs,
                       sizeof (FrameQ_ShMem_Attrs),
                       Cache_Type_ALL,
                       TRUE);
        }
        queue = obj->frmQueue[reader->frameQIndex];
        if (obj->ctrlStructCacheFlag  == TRUE) {
            Cache_inv((Ptr)queue,
                      sizeof(FrameQ_ShMem_FrameQueue),
                      Cache_Type_ALL,
                      TRUE);
        }
        for (i = 0; i < count; i++) {
            filledQueueNo = filledQueueId[i];
            /* Check if the filledquque has sufficient frames */
             if ( queue->numFrames[filledQueueNo] < pNumFrames [i]) {
                 status = FrameQ_E_INSUFFICENT_FRAMES;
                 /* We are using subNotifyentry 0 for framePool 0 and
                  * so on
                  */
                 pNotifySubIds[i] = filledQueueNo;
                 numSubEntries +=1;

             }
        }

        if (status >= 0) {
            for (i = 0; i < *numFrames; i++) {
                filledQueueNo = filledQueueId[i];
                if (queue->numFrames[filledQueueNo] == 0 ) {
                    /* Should not be here */
                    status = FrameQ_E_EMPTY;
                    pframe[i] = NULL;
                    if (    (reader->isRegistered == TRUE)
                         && (reader->notifyId != (UInt32)-1)) {
                        /* Enable notification on the subnotifyEntry  to
                         * get notification on this filledQueueId.
                         */
                        ClientNotifyMgr_enableNotification (
                                             (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                             reader->notifyId,
                                             filledQueueNo);
                    }
                }
                else {
                    frame = (FrameQ_ShMem_Frame)ListMP_getHead ((ListMP_Handle)
                        obj->frameList[reader->frameQIndex][filledQueueNo]);

                    Assert_isTrue ((NULL != frame),
                                   FrameQ_ShMem_A_nullPointer);
                    if (frame != NULL) {
                        queue->numFrames[filledQueueNo] -= 1;
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_wbInv ((Ptr)queue,
                                         sizeof (FrameQ_ShMem_FrameQueue),
                                         Cache_Type_ALL,
                                         TRUE);
                        }

                        /* Invalidate  frame header.
                         * Conevert portable frame pointers to virtual
                         * pointers.Then do invalidate frame buffers.
                         */
                        if (obj->frmHdrBufCacheFlag == TRUE){
                            status = FrameQBufMgr_invalidateHeaderBuf(
                                                    obj->frameQBufMgrHandle,
                                                    frame);
                        }

                        /* Convert  portable address of framebuffers to
                         * virtual address in the retrieved frame.
                         */
                        for ( j = 0; j < frame->numFrameBuffers; j++) {
                            /* Address translation to virtual pointers */
                            status =  FrameQBufMgr_translateAddr (
                                         obj->frameQBufMgrHandle,
                                         &bufPtr,
                                         FrameQBufMgr_AddrType_Virtual,
                                         (Ptr)frame->frameBufInfo[j].bufPtr,
                                         FrameQBufMgr_AddrType_Portable,
                                         FrameQBufMgr_BUF_FRAMEBUF);

                            Assert_isTrue ( (NULL != bufPtr),
                                             FrameQ_ShMem_A_nullPointer);
                            frame->frameBufInfo[j].bufPtr = (UInt32)bufPtr;

                            if (   (obj->frmBufCacheFlag[j] == TRUE)
                                && (obj->bufCpuAccessFlag[j] == TRUE)){
                                /* TODO: Invalidate  frame buffers pointed to by this
                                 * frame Using optimum cache coherence operations
                                 */
                                status = FrameQBufMgr_invalidateFrameBuf(
                                      obj->frameQBufMgrHandle,
                                      (Ptr)(frame->frameBufInfo[j].bufPtr +
                                      frame->frameBufInfo[j].startOffset),
                                      frame->frameBufInfo[j].validSize,
                                      j);
                             }
                        }
                        pframe[i] = frame;
                        lNumFrames += 1;
                    }
                }/*  else of queue->numFrames[filledQueueNo] == 0 */
            } /* for (i = 0; i < *numFrames; i++) */

            if ((lNumFrames > 0) && (lNumFrames < *numFrames)) {
                status = FrameQ_E_INSUFFICENT_FRAMES;
            }

        } /* if (status >= 0) */

        if (   (  (status == FrameQ_E_INSUFFICENT_FRAMES)
                ||(status == FrameQ_E_EMPTY))
             && (   (reader->isRegistered == TRUE)
                 && (reader->notifyId != (UInt32)-1))) {
            tmpStatus = ClientNotifyMgr_waitOnMultiSubEntryNotify(
                                             (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                             reader->notifyId,
                                             pNotifySubIds,
                                             numSubEntries);
            if (tmpStatus < 0) {
                status = FrameQ_E_FAIL;
            }
        }
    } /* if(reader->isValid != TRUE) {*/

    /* Release lock */
    GateMP_leave ((GateMP_Handle)obj->gate , key);

    if (lNumFrames ==0) {
        for (i = 0; i < *numFrames; i++) {
            pframe[i] = NULL;
        }
    }

    *numFrames = lNumFrames;

    return (status);
}

/*
 * ======== FrameQ_ShMem_dup ========
 * Function to duplicate the given frame
 */
Int32
FrameQ_ShMem_dup (FrameQ_ShMem_Object           *obj,
                  FrameQ_ShMem_Frame            frame,
                  FrameQ_ShMem_Frame*           dupedFrame)
{
    Int32                   status       = FrameQ_S_SUCCESS;
    FrameQBufMgr_Frame      dupedFramePtr[1];

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != dupedFrame),
                   FrameQ_ShMem_A_invalidParams);

    /* Allocate frame header for the duplicated frame */

    status = FrameQBufMgr_dup (obj->frameQBufMgrHandle,
                               (FrameQBufMgr_Frame)frame,
                               dupedFramePtr,
                               1);
    /* No Need to check status to assign dupedFramePtr.
     * dupedFramePtr[0] will be NULL if dup fails.
     */
    *dupedFrame = (FrameQ_ShMem_Frame) dupedFramePtr[0];

    return (status);
}

/*
 *  ======== FrameQ_ShMem_flushAll ========
 *  API to flush frames.
 *
 *  Flushes all the queus. Currently supported for creator.
 */
Int32
FrameQ_ShMem_flushAll (FrameQ_ShMem_Object *obj)
{
    Int32                        status  = FrameQ_S_SUCCESS;
    UInt32                       count =0;
    UInt32                      numFlushFrames = 0;
    volatile FrameQ_ShMem_FrameQueue     *queue;
    FrameQ_ShMem_Frame           frame;
    Ptr                          bufPtr;
    FrameQ_ShMem_Frame           pframe[FrameQ_MAX_FRAMESINVAPI];
    Int32                        key;
    UInt32                       i;
    UInt16                       j;
    UInt16                       k;
    UInt16                       m;
    Int32                        tmpStatus;

    UInt8                        filledQueueNo;

    Assert_isTrue ((NULL != obj),
                    FrameQ_ShMem_A_invalidParams );
    if (obj->objMode == FrameQ_MODE_NONE) {
        /* FrameQ creator  has called the flush.*/
        /* Flush  given queue from each reader */
        /* Acquire  lock */
        key = GateMP_enter ((GateMP_Handle)obj->gate);
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)obj->attrs,
                       sizeof (FrameQ_ShMem_Attrs),
                       Cache_Type_ALL,
                       TRUE);
        }
        for (k =0; k < obj->attrs->numQueues;k++) {
            filledQueueNo = k;
            for (j = 0; j < obj->attrs->numReaders; j++) {
                /* Each queu( obj->frmQueue) is for  one reader. loop through
                 * all the queues and free it
                 */
                queue = obj->frmQueue[j];
                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_inv ((Ptr)queue,
                               sizeof (FrameQ_ShMem_FrameQueue),
                               Cache_Type_ALL,
                               TRUE);
                }
                do {
                    count = queue->numFrames[filledQueueNo];
                    for (i = 0; i < FrameQ_MAX_FRAMESINVAPI && i < count ; i++) {
                        frame = (FrameQ_ShMem_Frame)ListMP_getHead ((ListMP_Handle)
                                    obj->frameList[j][filledQueueNo]);
                        if (frame != NULL) {
                            queue->numFrames[filledQueueNo] -= 1;
                            if (obj->ctrlStructCacheFlag == TRUE) {
                                Cache_wbInv ((Ptr)queue,
                                             sizeof (FrameQ_ShMem_FrameQueue),
                                             Cache_Type_ALL,
                                             TRUE);
                            }

                            /* Invalidate  frame header.
                             * Conevert portable frame pointers to virtual pointers.
                             * Then do invalidate frame buffers.
                             */
                            if (obj->frmHdrBufCacheFlag == TRUE){
                                status = FrameQBufMgr_invalidateHeaderBuf(
                                                            obj->frameQBufMgrHandle,
                                                            frame);
                            }
                            /* Convert  portable address of framebuffers to  virtual
                             * address in the retrieved frame.
                             */
                             for ( m = 0; m < frame->numFrameBuffers; m++) {
                                /* Address translation to virtual pointers */
                                status =  FrameQBufMgr_translateAddr (
                                                    obj->frameQBufMgrHandle,
                                                    &bufPtr,
                                                    FrameQBufMgr_AddrType_Virtual,
                                                    (Ptr)frame->frameBufInfo[m].bufPtr,
                                                    FrameQBufMgr_AddrType_Portable,
                                                    FrameQBufMgr_BUF_FRAMEBUF);

                                Assert_isTrue ( (NULL != bufPtr),
                                                 FrameQ_ShMem_A_nullPointer);
                                frame->frameBufInfo[m].bufPtr = (UInt32)bufPtr;

                                if (   (obj->frmBufCacheFlag[m] == TRUE)
                                   && (obj->bufCpuAccessFlag[m] == TRUE)) {
                                    /* Invalidate  frame buffers pointed to by this
                                     * frame Using optimum cache coherence operations
                                     */
                                    status = FrameQBufMgr_invalidateFrameBuf(
                                                    obj->frameQBufMgrHandle,
                                                   (Ptr)(frame->frameBufInfo[m].bufPtr +
                                                    frame->frameBufInfo[m].startOffset),
                                                    frame->frameBufInfo[m].validSize,
                                                    m);
                                }
                            }
                            pframe[i]= frame;
                            numFlushFrames ++;
                        }
                    }
                    if (numFlushFrames > 0) {
                        /* Leave the lock and Call FrameQBufMgr_freeV API.*/
                        GateMP_leave ((GateMP_Handle)obj->gate , key);
                        /* Call FrameQBufMgr API to free frame */
                        tmpStatus = FrameQBufMgr_freev(obj->frameQBufMgrHandle,
                                                       (FrameQBufMgr_Frame *)pframe,
                                                       numFlushFrames);
                        Assert_isTrue(tmpStatus >= 0, FrameQ_ShMem_A_apiFail);
                        if (tmpStatus < 0) {
                            status = FrameQ_E_FAIL;
                        }

                        key = GateMP_enter((GateMP_Handle)obj->gate);
                    }
                    /* Loop again if writer has put any frames  to
                     * total frames are more than pframe array size
                     */
                    count = queue->numFrames[filledQueueNo];
                } while (count > 0 );
            }
        }
        /* Release lock */
        GateMP_leave ((GateMP_Handle)obj->gate , key);
    }
    return status;
}

/*
 *  ======== FrameQ_ShMem_registerNotifier ========
 *  API to resister a call back function.
 *
 * For reader :
 * It registers notification function to notify about available frames in the
 * FrameQ.
 * For writer :
 * It registers notification function to notify about available free frames in the
 * Plugged in FrameQBufMgr.
 */
Int32
FrameQ_ShMem_registerNotifier (FrameQ_ShMem_Object *obj,
                               FrameQ_ShMem_NotifyParams *notifyParams)
{
    Int32                        status       = FrameQ_S_SUCCESS;
    ClientNotifyMgr_registerNoitifyParams  regParams;
    FrameQBufMgr_NotifyParams    lnotifyParams;
    Int32                        tmpStatus;
    UInt32                       key;
    FrameQ_ShMem_ReaderClient   *reader;
    FrameQ_ShMem_WriterClient   *writer;
    UInt32                      lNotifyId;

    Assert_isTrue ((NULL != obj), FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != notifyParams), FrameQ_ShMem_A_invalidParams);

    switch (obj->objMode) {

        case FrameQ_ShMem_MODE_READER:
        {
            reader = (FrameQ_ShMem_ReaderClient*)obj->virtClientHandle;
            if (obj->ctrlStructCacheFlag == TRUE) {
                key = GateMP_enter ((GateMP_Handle)obj->gate);
                Cache_inv (reader,
                           sizeof (FrameQ_ShMem_ReaderClient),
                           Cache_Type_ALL,
                           TRUE);
                GateMP_leave ((GateMP_Handle)obj->gate, key);
            }
            regParams.clientHandle   = notifyParams->cbClientHandle;
            regParams.procId         = MultiProc_self();
            regParams.notifyType     = notifyParams->notifyType;
            regParams.fxnPtr         = (ClientNotifyMgr_FnCbck)notifyParams->cbFxnPtr;
            regParams.cbContext      = notifyParams->cbContext;
            regParams.watermarkCond1 = notifyParams->watermark;
            regParams.watermarkCond2 = (UInt32)-1;

            tmpStatus = ClientNotifyMgr_registerClient((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                       &regParams,
                                                       (Ptr)&lNotifyId);
            if (tmpStatus < 0) {
                status = FrameQ_E_CLIENTNOTIFYMGRREGCLIENT;
            }
            else {
                reader->notifyId = lNotifyId;
                reader->isRegistered = TRUE;
                if (obj->ctrlStructCacheFlag == TRUE) {
                    key = GateMP_enter ((GateMP_Handle)obj->gate);
                    Cache_wbInv (reader,
                                 sizeof (FrameQ_ShMem_ReaderClient),
                                 Cache_Type_ALL,
                                 TRUE);
                    GateMP_leave ((GateMP_Handle)obj->gate, key);
                }
           }
        }
        break;

        case FrameQ_ShMem_MODE_WRITER:
        {
            writer = (FrameQ_ShMem_WriterClient *) obj->writer;
            if (obj->ctrlStructCacheFlag == TRUE) {
                key = GateMP_enter ((GateMP_Handle)obj->gate);
                Cache_inv (writer,
                           sizeof (FrameQ_ShMem_WriterClient),
                           Cache_Type_ALL,
                           TRUE);
                GateMP_leave ((GateMP_Handle)obj->gate, key);
            }

            lnotifyParams.cbClientHandle = notifyParams->cbClientHandle;
            lnotifyParams.notifyType     = (SysLink_NotifyType)notifyParams->notifyType;
            lnotifyParams.cbFxnPtr       = (FrameQBufMgr_NotifyFunc)
                                                         notifyParams->cbFxnPtr;
            lnotifyParams.cbContext      = notifyParams->cbContext;
            lnotifyParams.watermark      = notifyParams->watermark;

            tmpStatus = FrameQBufMgr_registerNotifier(obj->frameQBufMgrHandle,
                                                     &lnotifyParams);
            if (tmpStatus < 0) {
                status = FrameQ_E_FRAMEQBUFMGRREGCLIENT;
            }
            else {
                writer->isRegistered = TRUE;
                if (obj->ctrlStructCacheFlag == TRUE) {
                    key = GateMP_enter ((GateMP_Handle)obj->gate);
                    Cache_wbInv (writer,
                                 sizeof (FrameQ_ShMem_WriterClient),
                                 Cache_Type_ALL,
                                 TRUE);
                    GateMP_leave ((GateMP_Handle)obj->gate, key);
                }
           }
        }
        break;

        default:
            /* Invalid mode. Return error */
            status = FrameQ_E_ACCESSDENIED;
        break;
    }

    return (status);
}

/*
 *  ======== FrameQ_ShMem_unregisterNotifier ========
 *  API to unresister a call back function.
 *
 */
Int32
FrameQ_ShMem_unregisterNotifier (FrameQ_ShMem_Object *obj)
{
    Int32                        status       = FrameQ_S_SUCCESS;
    Int32                        tmpStatus;
    UInt32                       key;
    FrameQ_ShMem_ReaderClient   *reader;
    FrameQ_ShMem_WriterClient   *writer;

    Assert_isTrue ((NULL != obj), FrameQ_ShMem_A_invalidParams);

    switch (obj->objMode) {

        case FrameQ_ShMem_MODE_READER:
        {
            /* For Reader  unregister the call back with the client notify
             * manager.
             */
            reader = (FrameQ_ShMem_ReaderClient*)obj->virtClientHandle;
            if (obj->ctrlStructCacheFlag == TRUE) {
                key = GateMP_enter ((GateMP_Handle)obj->gate);
                Cache_inv (reader,
                           sizeof (FrameQ_ShMem_ReaderClient),
                           Cache_Type_ALL,
                           TRUE);
                GateMP_leave ((GateMP_Handle)obj->gate, key);
            }
            tmpStatus = ClientNotifyMgr_unregisterClient (
                                                     (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                     reader->notifyId);
            if (tmpStatus < 0) {
                status = FrameQ_E_CLIENTNOTIFYMGRUNREGCLIENT;
            }
            else {
                reader->isRegistered = FALSE;
                reader->notifyId     = (UInt32)-1;
                if (obj->ctrlStructCacheFlag == TRUE) {
                    key = GateMP_enter ((GateMP_Handle)obj->gate);
                    Cache_wbInv (reader,
                                 sizeof (FrameQ_ShMem_ReaderClient),
                                 Cache_Type_ALL,
                                 TRUE);
                    GateMP_leave ((GateMP_Handle)obj->gate, key);
                }
            }
        }
        break;

        case FrameQ_ShMem_MODE_WRITER:
        {
            /* For writer  unregister the call back with the FrameQBufMgr.
             */
            writer = (FrameQ_ShMem_WriterClient *) obj->writer;
            if (obj->ctrlStructCacheFlag == TRUE) {
                key = GateMP_enter ((GateMP_Handle)obj->gate);
                Cache_inv (writer,
                           sizeof (FrameQ_ShMem_ReaderClient),
                           Cache_Type_ALL,
                           TRUE);
                GateMP_leave ((GateMP_Handle)obj->gate, key);
            }

            tmpStatus = FrameQBufMgr_unregisterNotifier(obj->frameQBufMgrHandle);
            if (tmpStatus < 0) {
                status = FrameQ_E_FRAMEQBUFMGRUNREGCLIENT;
            }
            else {
                writer->isRegistered = FALSE;
                if (obj->ctrlStructCacheFlag == TRUE) {
                    key = GateMP_enter ((GateMP_Handle)obj->gate);
                    Cache_wbInv (writer,
                                 sizeof (FrameQ_ShMem_WriterClient),
                                 Cache_Type_ALL,
                                 TRUE);
                    GateMP_leave ((GateMP_Handle)obj->gate, key);
                }
           }
        }
        break;

        default:
            /* Invalid mode. Return error */
            status = FrameQ_E_ACCESSDENIED;
        break;
    }

    return (status);
}

/*
 * ======== FrameQ_ShMem_sendNotify ========
 * Function to send forced notification to the reader clients.
 */
Int32
FrameQ_ShMem_sendNotify (FrameQ_ShMem_Object * obj, UInt16 msg)
{
    Int32   status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != obj), FrameQ_ShMem_A_invalidParams);

    switch (obj->objMode) {

        case FrameQ_ShMem_MODE_WRITER:
        {
            /* Send force notification to all the registered Reader clients.
             * status  will be  success if atleast one reader is notified.
             */
            status = ClientNotifyMgr_sendForceNotificationMulti(
                                                    (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                    (UInt32)-1,
                                                     msg);
        }
        break;

        default:
            /* Invalid mode. Return error */
            status = FrameQ_E_ACCESSDENIED;
        break;
    }

    return (status);
}

/*
 *  ======== FrameQ_ShMem_getNumFrames ========
 *  API to find out the number of frames in a FrameQ .
 *  For Writer:
 *        It returns the  number of frames in the primary reader's filled
 *        queue 0.
 *  For Reader:
 *       It returns the number of frames in the Reader's(caller)filledqueue 0.
 */
Int32
FrameQ_ShMem_getNumFrames(FrameQ_ShMem_Object *obj,
                          UInt32         *numFrames)
{
    Int32                            status  = FrameQ_S_SUCCESS;
    volatile FrameQ_ShMem_Attrs          *pAttrs;
    FrameQ_ShMem_ReaderClient            *reader;
    volatile FrameQ_ShMem_FrameQueue     *queue;
    UInt32                                queueId;
    UInt32                                key;
    UInt8                                 filledQueueNo;

    Assert_isTrue ((NULL != obj), FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFrames), FrameQ_ShMem_A_invalidParams);

    pAttrs = obj->attrs;
    /* This function returns  number of frames available in the fileld queue 0
     */
    filledQueueNo = 0;

    switch (obj->objMode)
    {
        case FrameQ_ShMem_MODE_READER:
        {
             key = GateMP_enter ((GateMP_Handle)obj->gate);
             reader  = obj->virtClientHandle;
            /* Return  available frames in the frame queue associated withe this
             * reader  excluding the frames that it has already acquired but not
             * freed.
             */
            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)pAttrs,
                          sizeof(FrameQ_ShMem_Attrs),
                          Cache_Type_ALL,
                          TRUE);
                Cache_inv(reader,
                          sizeof(FrameQ_ShMem_ReaderClient),
                          Cache_Type_ALL,
                          TRUE);
            }
            queueId = reader->frameQIndex;
            queue = obj->frmQueue[queueId];
            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)queue,
                          sizeof(FrameQ_ShMem_FrameQueue),
                          Cache_Type_ALL,
                          TRUE);
            }
            /* Get valid frames  */
            *numFrames = queue->numFrames[filledQueueNo];

            GateMP_leave ((GateMP_Handle)obj->gate, key);
        }
        break;

        case FrameQ_ShMem_MODE_WRITER:
        {
            key = GateMP_enter ((GateMP_Handle)obj->gate);

            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)pAttrs,
                          sizeof(FrameQ_ShMem_Attrs),
                          Cache_Type_ALL,
                          TRUE);
            }
            queueId = pAttrs->primaryQueueId;
            queue = obj->frmQueue[queueId];
            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)queue,
                          sizeof(FrameQ_ShMem_FrameQueue),
                          Cache_Type_ALL,
                          TRUE);
            }
            /* Get valid frames  */
            *numFrames = queue->numFrames[filledQueueNo];

            GateMP_leave ((GateMP_Handle)obj->gate, key);
        }
        break;

        default:
        {
            *numFrames = 0;
             status    = FrameQ_E_FAIL;
        }
    }

    return (status);
}

/*
 *  ======== FrameQ_getvNumFrames ========
 * For Reader this function is to get the available frames  from the queues
 * associated with this reader client.
 * For writer this function is to get the available frames  from the queues
 * associated with the primary Reader(first reader).
 */
Int32
FrameQ_ShMem_getvNumFrames (FrameQ_ShMem_Object   * obj,
                            UInt32                  numFrames[],
                            Uint8                   filledQId[],
                            UInt8                   numFilledQids)
{
    Int32       status = FrameQ_S_SUCCESS;
    volatile FrameQ_ShMem_Attrs          *pAttrs;
    FrameQ_ShMem_ReaderClient            *reader;
    volatile FrameQ_ShMem_FrameQueue     *queue;
    UInt32      queueId;
    UInt32      i;
    Int32       key;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFrames),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != filledQId),
                   FrameQ_ShMem_A_invalidParams);

    pAttrs = obj->attrs;
    /* This function returns  number of frames available in the fileld queue 0
     */
    switch (obj->objMode)
    {
        case FrameQ_ShMem_MODE_READER:
        {
            key = GateMP_enter ((GateMP_Handle)obj->gate);
            reader  = obj->virtClientHandle;
            /* Return  available frames in the frame queue associated withe this
             * reader  excluding the frames that it has already acquired but not
             * freed.
             */
            if (obj->ctrlStructCacheFlag  == TRUE) {
                    Cache_inv((Ptr)pAttrs,
                               sizeof(FrameQ_ShMem_Attrs),
                               Cache_Type_ALL,
                               TRUE);
                    Cache_inv(reader,
                              sizeof(FrameQ_ShMem_ReaderClient),
                              Cache_Type_ALL,
                              TRUE);
            }
            queueId = reader->frameQIndex;
            queue = obj->frmQueue[queueId];
            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)queue,
                          sizeof(FrameQ_ShMem_FrameQueue),
                          Cache_Type_ALL,
                          TRUE);
            }
            for (i = 0; i < numFilledQids; i++) {
                if (filledQId[i] > pAttrs->numQueues) {
                    status = FrameQ_E_INVALIDARG;
                }
                else {
                     /* Get valid frames  */
                    numFrames[i] = queue->numFrames[filledQId[i]];
                }

            }
            GateMP_leave ((GateMP_Handle)obj->gate, key);
        }
        break;

        case FrameQ_ShMem_MODE_WRITER:
        {
            key = GateMP_enter ((GateMP_Handle)obj->gate);

            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)pAttrs,
                          sizeof(FrameQ_ShMem_Attrs),
                          Cache_Type_ALL,
                          TRUE);
            }
            queueId = pAttrs->primaryQueueId;
            queue = obj->frmQueue[queueId];
            if (obj->ctrlStructCacheFlag  == TRUE) {
                Cache_inv((Ptr)queue,
                          sizeof(FrameQ_ShMem_FrameQueue),
                          Cache_Type_ALL,
                          TRUE);
            }
            /* Get valid frames  */
            for (i = 0; i < numFilledQids; i++) {
                if (filledQId[i] > pAttrs->numQueues) {
                    status = FrameQ_E_INVALIDARG;
                }
                else {
                     /* Get valid frames  */
                    numFrames[i] = queue->numFrames[filledQId[i]];
                }
            }
            GateMP_leave ((GateMP_Handle)obj->gate, key);
        }
        break;

        default:
        {
            for (i = 0; i < numFilledQids; i++) {
                 /* Get valid frames  */
                numFrames[i] = 0;

            }
            status    = FrameQ_E_FAIL;
        }
    }


    return (status);
}

/*
 *  ======== FrameQ_ShMem_getNumFreeFrames ========
 *  Function to get the number of free frames available in the plugged in
 *  FrameQbufMgr's FreeFramePool 0. The returned number free frame may not
 *  exist in the frameQbufMgr  after this call.This is because
 *  if any allocs are done by the other FrameQ writers on the
 *  same FrameQBufMgr.If frees are not other frameQ writers the free Frames
 *  would be even more than the value that we have got.
 */
Int32
FrameQ_ShMem_getNumFreeFrames(FrameQ_ShMem_Object *obj,
                              UInt32              *numFreeFrames)
{
    Int32                            status  = FrameQ_S_SUCCESS;
    Int32                            tmpStatus;

    Assert_isTrue ((NULL != obj), FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFreeFrames), FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != obj->frameQBufMgrHandle),
                    FrameQ_ShMem_A_invalidParams);

    tmpStatus = FrameQBufMgr_getNumFreeFrames (obj->frameQBufMgrHandle,
                                               numFreeFrames);
    if (tmpStatus < 0) {
        status = tmpStatus;
    }

    return (status);
}

/*
 *  ======== FrameQ_getvNumFreeFrames ========
 * Function to get the number of free frames available in the plugged in
 * FrameQbufMgr.
 */
Int32
FrameQ_ShMem_getvNumFreeFrames (FrameQ_ShMem_Object   * obj,
                                UInt32                  numFreeFrames[],
                                UInt8                   freeQId[],
                                UInt8                   numFreeQids)
{
    Int32       status = FrameQ_S_SUCCESS;
    Int32       tmpStatus;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFreeFrames),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != freeQId),
                   FrameQ_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != obj->frameQBufMgrHandle),
                    FrameQ_ShMem_A_invalidParams);

    tmpStatus = FrameQBufMgr_getvNumFreeFrames (obj->frameQBufMgrHandle,
                                                numFreeFrames,
                                                freeQId,
                                                numFreeQids);
    if (tmpStatus < 0) {
        status = tmpStatus;
    }

    return (status);
}

/*
 *  ======== FrameQ_ShMem_control ========
 *  @brief  Provides a hook to perform implementation dependent operation.
 *
 */
Int32
FrameQ_ShMem_control (FrameQ_ShMem_Object *obj,
                      Int32                cmd,
                      Ptr                  arg)
{
    Int32   status = FrameQ_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQ_ShMem_A_invalidParams);

     status = FrameQ_E_NOTIMPLEMENTED;

     return (status);
}


/*
 *  ======== FrameQ_ShMem_getNumFrameBuffers ========
 *  API to find out the number of frame buffers in a frame.
 */
UInt32
FrameQ_ShMem_getNumFrameBuffers (FrameQ_ShMem_Frame frame )
{
    UInt32 numbufs = 0;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);

    numbufs =  frame->numFrameBuffers;

    return (numbufs);
}

/*
 *  ======== FrameQ_ShMem_getFrameBuffer ========
 *  API to retrieve framebuffer  identified by frameBufNum from a given frame.
 * Here no validation is performed for frameBufNum. User is expected to pass the
 * right value for it.
 * ( APPs can use getNumFrameBuffers API to know the number of frame buffers
 *  associated with the frame)
 */
Ptr
FrameQ_ShMem_getFrameBuffer( FrameQ_ShMem_Frame frame, UInt32 frameBufNum )
{
    Void * frmBuffer = NULL;
    FrameQ_ShMem_FrameBufInfo *frameBufInfo;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);
    Assert_isTrue ((frame->numFrameBuffers >= frameBufNum),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo = (FrameQ_ShMem_FrameBufInfo *)&(frame->frameBufInfo[0]);

    frmBuffer = (Ptr) frameBufInfo[frameBufNum].bufPtr;

    return (frmBuffer);
}

/*
 *  ======== FrameQ_ShMem_getFrameBufSize ========
 *  API to get the size of the framebuffer.
 */
UInt32
FrameQ_ShMem_getFrameBufSize(FrameQ_ShMem_Frame frame, UInt32 frameBufNum )
{
    UInt32 size;
    FrameQ_ShMem_FrameBufInfo *frameBufInfo;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);
    Assert_isTrue ((frame->numFrameBuffers >= frameBufNum),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo = (FrameQ_ShMem_FrameBufInfo *)&(frame->frameBufInfo[0]);
    size = frameBufInfo[frameBufNum].bufSize;

    return (size);
}


/*
 *  ======== FrameQ_ShMem_getFrameBufValidSize ========
 *  API to get the valid data size of a framebuffer  identified by frameBufNum
 *  in a given frame.
 */
UInt32
FrameQ_ShMem_getFrameBufValidSize(FrameQ_ShMem_Frame frame, UInt32 frameBufNum )
{
    UInt32 size;
    FrameQ_ShMem_FrameBufInfo *frameBufInfo;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);
    Assert_isTrue ((frame->numFrameBuffers >= frameBufNum),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo = (FrameQ_ShMem_FrameBufInfo *)&(frame->frameBufInfo[0]);
    size = frameBufInfo[frameBufNum].validSize;

    return (size);
}

/*
 *  ======== FrameQ_ShMem_getFrameBufDataStartOffset ========
 *  API to get the valid data start offset in framebuffer  identified by
 *  frameBufNum in a given frame.
 */
UInt32
FrameQ_ShMem_getFrameBufDataStartOffset(FrameQ_ShMem_Frame frame,
                                        UInt32         frameBufNum )
{
    UInt32 offSet;
    FrameQ_ShMem_FrameBufInfo *frameBufInfo;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);
    Assert_isTrue ((frame->numFrameBuffers >= frameBufNum),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo = (FrameQ_ShMem_FrameBufInfo *)&(frame->frameBufInfo[0]);
    offSet = frameBufInfo[frameBufNum].startOffset;

    return (offSet);
}

/*
 *  ======== FrameQ_ShMem_setFrameBufValidSize ========
 *  API to set the valid data size of a framebuffer  identified by
 *  frameBufNum in  given frame.
 */
Int32
FrameQ_ShMem_setFrameBufValidSize(FrameQ_ShMem_Frame frame,
                                  UInt32             frameBufNum,
                                  UInt32             validDataSize )
{
    Int32                      status       = FrameQ_S_SUCCESS;
    FrameQ_ShMem_FrameBufInfo *frameBufInfo;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);
    Assert_isTrue ((frame->numFrameBuffers > frameBufNum),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo = (FrameQ_ShMem_FrameBufInfo *)&(frame->frameBufInfo[0]);

    Assert_isTrue ((frameBufInfo[frameBufNum].bufSize <
                       (validDataSize + frameBufInfo[frameBufNum].startOffset)),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo[frameBufNum].validSize = validDataSize;

    return (status);
}

/*
 *  ======== FrameQ_ShMem_setFrameBufDataStartOffset ========
 *  API to set the valid data start offset of a framebuffer  identified by
 *  frameBufNum in a given frame.
 */
Int32
FrameQ_ShMem_setFrameBufDataStartOffset (FrameQ_ShMem_Frame frame,
                                         UInt32             frameBufNum,
                                         UInt32             dataStartOffset)
{
    Int32                      status       = FrameQ_S_SUCCESS;
    FrameQ_ShMem_FrameBufInfo *frameBufInfo;

    Assert_isTrue ((NULL != frame), FrameQ_ShMem_A_invalidParams);
    Assert_isTrue ((frame->numFrameBuffers > frameBufNum),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo = (FrameQ_ShMem_FrameBufInfo *)&(frame->frameBufInfo[0]);

    Assert_isTrue ((frameBufInfo[frameBufNum].bufSize <
                       (dataStartOffset + frameBufInfo[frameBufNum].validSize)),
                    FrameQ_ShMem_A_invalidParams);

    frameBufInfo[frameBufNum].startOffset = dataStartOffset;

    return (status);
}


/*
 * ========================== Internal functions ============================
 */

/*
 *  ======== FrameQ_ShMem_Module_startup ========
 */
Int FrameQ_ShMem_Module_startup(Int phase)
{
    Int32   key;

    if (FrameQ_ShMem_module->nameServer != NULL) {
        if (!ti_sdo_utils_NameServer_Module_startupDone()) {
            return (Startup_NOTDONE);
        }
    }

    /* Check the  Front end module setup is done. */
    if (!FrameQ_Module_startupDone()) {
        return (Startup_NOTDONE);
    }

    key = Gate_enterSystem();

    FrameQ_ShMem_module->refCount++;

    if(FrameQ_ShMem_module->refCount > 1) {
        Gate_leaveSystem(key);
        return (Startup_DONE);
    }
    else {
        /*Open the Name server created by FrameQ front end*/
        if (FrameQ_ShMem_module->nameServer == NULL) {
            /* Open  the Name Server instance  created by Frond End FrameQBufMgr
             * module if not  assigned statically.
             */
//          FrameQ_ShMem_module->nameServer =
//                                  NameServer_getHandle(FRAMEQ_NAMESERVERNAME);
            FrameQ_ShMem_module->nameServer = gFrameQ_nsHandle ;
            if(NULL == FrameQ_ShMem_module->nameServer) {

                /* Decrement the reference count */
                FrameQ_ShMem_module->refCount--;
                Gate_leaveSystem(key);
                return (Startup_NOTDONE);
            }
        }
    }/* if(FrameQ_ShMem_module->refCount > 1) { */

    /* Release lock */
    Gate_leaveSystem(key);

    return (Startup_DONE);
}
/*
 *  ======== FrameQ_Shmem_updateCacheFlags ========
 *  Internal API to populate the cache flags
 */
Void FrameQ_Shmem_updateCacheFlags(FrameQ_ShMem_Object *obj,
                                   UInt32               cpuAccessFlags)
{
    UInt32  i = 0;

    if ( FrameQBufMgr_isCacheEnabledForHeaderBuf(obj->frameQBufMgrHandle) == TRUE) {
        obj->frmHdrBufCacheFlag = TRUE;
    }
    else {
        obj->frmHdrBufCacheFlag = FALSE;
    }

    for (i = 0; i < FrameQBufMgr_MAX_FRAMEBUFS ; i++) {
        /* Check and update if user has provided cache  flags for Frame
         * buffers.
         */
        if ( FrameQBufMgr_isCacheEnabledForFrameBuf(obj->frameQBufMgrHandle, i) == TRUE) {
            obj->frmBufCacheFlag[i] = TRUE;
        }
        else {
            obj->frmBufCacheFlag[i] = FALSE;
        }
        /* Check and update if user has provided cpu access flags for Frame
         * buffers.
         */
        if (   cpuAccessFlags
             & (1 << (FrameQ_ShMem_FRAMEBUF_CPUACCESSFLAGS_BITOFFSET +i))) {
            obj->bufCpuAccessFlag[i] = TRUE;
        }
        else {
            obj->bufCpuAccessFlag[i] = FALSE;
        }
    }
}

/*
 *  ======== FrameQ_ShMem_setOpenFlags ========
 *  API to  set the mode for the created client.
 */
Int32
FrameQ_ShMem_setOpenFlags (FrameQ_ShMem_Object     *obj,
                           ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_OpenParams *openParams)
{
    Int32                   status      = FrameQ_S_SUCCESS;
    volatile FrameQ_ShMem_Attrs      *pAttrs;
    volatile FrameQ_ShMem_WriterClient *writer;
    volatile FrameQ_ShMem_ReaderClient *reader;
    volatile FrameQ_ShMem_FrameQueue   *frmQueue;
    UInt32                  key;
    UInt32                  i ;
    UInt32                  index;
    UInt32                  primQId;

     if ( (NULL == obj) || ( NULL == openParams) ||(obj->gate == NULL)) {
        status = FrameQ_E_INVALIDARG;
    }

    if (status >= 0) {
        pAttrs = obj->attrs;
        writer = obj->writer;
        /*Invalidate  Control */
        key = GateMP_enter ((GateMP_Handle)obj->gate);
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv((Ptr)pAttrs,
                       sizeof(FrameQ_ShMem_Attrs),
                       Cache_Type_ALL,
                       TRUE);
            Cache_inv((Ptr)writer,
                       sizeof(FrameQ_ShMem_WriterClient),
                       Cache_Type_ALL,
                       TRUE);

        }

        if (openParams->commonOpenParams.openMode == FrameQ_ShMem_MODE_WRITER) {
            /* Checking if writer is already  opened  */
            if (writer->isValid == TRUE) {
                status = FrameQ_E_WRITER_EXISTS;
            }
            else {
                writer->isValid = TRUE;
                /* Get the caller processor id from the MultiProc module */
                writer->procId  = MultiProc_self();

                writer->cacheFlags = openParams->commonOpenParams.cpuAccessFlags;
                /*TODO: Add cache flags for controlstruct and frame header buffer in writer->cacheFlags */

                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_wbInv ((Ptr)writer,
                                 sizeof (FrameQ_ShMem_WriterClient),
                                 Cache_Type_ALL,
                                 TRUE);
                }

                obj->virtClientHandle = (void*)writer;
                obj->objMode          = FrameQ_ShMem_MODE_WRITER;
            }

        }
        else if (   openParams->commonOpenParams.openMode
                 == FrameQ_ShMem_MODE_READER) {
            /* Check in  Reader clients to know which one is free */
            for (i = 0; i < pAttrs->numReaders; i++) {
                reader = obj->reader[i];
                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_inv((Ptr)reader,
                              sizeof(FrameQ_ShMem_ReaderClient),
                              Cache_Type_ALL,
                              TRUE);
                }
                if (reader->isValid != TRUE) {
                    break;
                }
            }

            if ( i ==  pAttrs->numReaders) {
                /* Empty slot is not found in the array of readers
                 * No more  additional readers*/
                status = FrameQ_E_MAX_READERS;
            }
            else {
                 /* Unused slot found in the reader list */
                 reader->isValid    = TRUE;
                 reader->procId     = MultiProc_self();
                 reader->cacheFlags =
                                    openParams->commonOpenParams.cpuAccessFlags;
                 pAttrs->numActiveReaders++;

                 if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_wbInv ((Ptr)pAttrs,
                                 sizeof (FrameQ_ShMem_Attrs),
                                 Cache_Type_ALL,
                                 TRUE);
                 }

                /* If it is the first reader and writer is still not active
                 * set  queue of this reader as primay queue
                 */
                if (pAttrs->numActiveReaders == 1) {
                    primQId = pAttrs->primaryQueueId;

                    frmQueue = obj->frmQueue[primQId];
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_inv ((Ptr)frmQueue,
                                   sizeof (FrameQ_ShMem_FrameQueue),
                                   Cache_Type_ALL,
                                   TRUE);
                    }

                    /*Check if  writer is in active state */
                    if(writer->isValid == TRUE) {
                        /* No need to change the primary queue
                         * instead point this reader to the  primary
                         * queque.
                         */
                        reader->frameQIndex = primQId;
                        frmQueue->isAllocated = TRUE;
                        frmQueue->readerIndex = i;
                    }
                    else {
                        /* If writer is not active,
                         * Then also point this reader to the primary queue
                         */
                        reader->frameQIndex = primQId;
                        frmQueue->isAllocated = TRUE;
                        frmQueue->readerIndex = i;
                    }

                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_wbInv ((Ptr)reader,
                                     sizeof (FrameQ_ShMem_ReaderClient),
                                     Cache_Type_ALL,
                                     TRUE);
                        Cache_wbInv ((Ptr)frmQueue,
                                     sizeof (FrameQ_ShMem_FrameQueue),
                                     Cache_Type_ALL,
                                     TRUE);
                    }
                }
                else {
                    /* Assign empty queue for this reader */
                    for (index = 0;
                         index < pAttrs->numReaders;
                         index++) {
                        frmQueue = obj->frmQueue[index];
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_inv ((Ptr)frmQueue,
                                       sizeof (FrameQ_ShMem_FrameQueue),
                                       Cache_Type_ALL,
                                       TRUE);
                        }

                        if (frmQueue->isAllocated != TRUE)
                        {
                            break;
                        }
                    }

                    if (index == pAttrs->numReaders) {
                        status = FrameQ_E_FAIL;
                    }
                    else {
                        reader->frameQIndex             =  index;
                        frmQueue->isAllocated = TRUE;
                        frmQueue->readerIndex = i;

                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_wbInv ((Ptr)reader,
                                         sizeof (FrameQ_ShMem_ReaderClient),
                                         Cache_Type_ALL,
                                         TRUE);
                            Cache_wbInv ((Ptr)frmQueue,
                                         sizeof (FrameQ_ShMem_FrameQueue),
                                         Cache_Type_ALL,
                                         TRUE);
                        }

                    }
                }
                if (status >= 0) {
                    /* Point to the obj->virtControlHandle to this reader */
                    obj->virtClientHandle = (void*)reader;
                    obj->objMode          = FrameQ_ShMem_MODE_READER;
                }
            }

        }
        else {
            /*Invalid mode specified.Return error */
            status = FrameQ_E_FAIL;
        }

        GateMP_leave ((GateMP_Handle)obj->gate, key);
    }

    return (status);
}

/*
 *  ======== _FrameQBufMgr_ShMem_create ========
 *  API to create/open the instance.Called internally by
 *  the FrameQBufMgr_create call.
 */
static inline
Int32
_FrameQ_ShMem_create (FrameQ_ShMem_Object *obj,
                      const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params *params,
                      Error_Block *eb)
{
    Int32                   status      = FrameQ_S_SUCCESS;
    UInt32                  virtAddr    = 0;
    UInt32                  addrOffset   = 0;
    Int32                   tmpStatus;
    volatile FrameQ_ShMem_Attrs  *pAttrs;
    volatile FrameQ_ShMem_WriterClient  * writer;
    ListMP_Params           listMPParams;
    UInt32                  i;
    SharedRegion_SRPtr      sharedShmBase;
    GateMP_Params           gateParams;
    GateMP_Handle           gateHandle;
    ClientNotifyMgr_Params  clientMgrParams;
    ClientNotifyMgr_Handle  cliMgrHandle;
    FrameQ_NameServerEntry  nameserverEntry;
    FrameQ_NameServerEntry  *pnameserverEntry;
    UInt32                  k;
    FrameQBufMgr_OpenParams frmQBufMgrOpenParams;
    FrameQBufMgr_Handle     bufMgrHandle;
    UInt32                  listMPsharedAddrSize;
    UInt32                  gateMPsharedAddrSize;
    Int32                   index;
    UInt32                  len;
    FrameQ_ShMem_Params     shparams;
//    UInt32                  key;

    Assert_isTrue ((NULL != params),FrameQ_ShMem_A_invalidParams);

    /* Assert if shared addr  provided is null  and region id provide is
     * invalid.
     */
    Assert_isTrue (    (params->sharedAddr != NULL)
                    || (params->regionId != SharedRegion_INVALIDREGIONID),
                    FrameQ_ShMem_A_invalidParams);

    /* Check if name   exists in the  nameServer  already */
    if (strlen((String)params->commonCreateParams.name) > 0) {
        /* Ensure NameServer is enabled */
        len = sizeof (FrameQ_NameServerEntry);
        status = NameServer_get ((NameServer_Handle)FrameQ_ShMem_module->nameServer,
                                (String)params->commonCreateParams.name,
                                &nameserverEntry,
                                &len ,
                                NULL);
        if (status > 0) {
            /* Already instance with the name exists */
            status = FrameQ_E_INST_EXISTS;
            Error_raise(eb,
                        FrameQ_ShMem_E_FrameQCreated,
                        NULL,
                        0);
        }
        else if (status == NameServer_E_FAIL) {
            /* Error happened in NameServer. Pass the error up. */
            status = FrameQ_E_FAIL;
        }
        else  {
            /* Instance with this name is not there in the NameServer.
             * So create can proceed.
             */
             /* Reinitialize the error block */
             Error_init(eb);
             status = FrameQ_S_SUCCESS;
        }
    }

    if (status >= 0) {
        /* Validate the input parameters */
        /* Check  if sufficient memory provided */
//        if (   (   params->sharedAddr != NULL)
//                && (   params->sharedAddrSize
//                     < FrameQ_ShMem_sharedMemReq (
//                                        (FrameQ_ShMem_Params*)params)
//                )) {
//              status = FrameQ_E_INVALIDARG;
//              Error_raise(eb,
//                          FrameQ_ShMem_E_InvSize,
//                          NULL,
//                          0);
//        }
//        else
        if (params->numReaders > FrameQ_MAX_INST_READERS ) {
              status = FrameQ_E_INVALIDARG;
              Error_raise(eb,
                          FrameQ_ShMem_E_InvalidArg,
                          NULL,
                          0);
        }
        else if (   (params->frameQBufMgrName == NULL)
                 && (params->frameQBufMgrSharedAddr == NULL)) {
              status = FrameQ_E_INVALIDARG;
              Error_raise(eb,
                          FrameQ_ShMem_E_InvalidArg,
                          NULL,
                          0);
        }
    }

    if (status >= 0) {
        if (params->sharedAddr == NULL) {
            /* Creating using a shared region ID and name */
            FrameQ_ShMem_getSharedParams (&shparams, params);
            /* Need to allocate from the heap */
            obj->allocSize = FrameQ_ShMem_sharedMemReq (&shparams);

            /* HeapMemMP will do the alignment */
            virtAddr = (UInt32)Memory_alloc(
                                            SharedRegion_getHeap(params->regionId),
                                            obj->allocSize,
                                            0,
                                            eb);
        }
        else {
            virtAddr = (UInt32)params->sharedAddr;
        }

        /* Creating using sharedAddr */
        obj->regionId = SharedRegion_getId((Ptr)virtAddr);

        /* Assert that the buffer is in a valid shared region */
        Assert_isTrue(obj->regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_addrNotInSharedRegion);

        /* Assert that sharedAddr is cache aligned */
        Assert_isTrue(((UInt32)virtAddr %
                      SharedRegion_getCacheLineSize(obj->regionId) == 0),
                      Ipc_A_addrNotCacheAligned);

        /* set object's cacheEnabled, objType, and attrs  */
        obj->ctrlStructCacheFlag =
                         SharedRegion_isCacheEnabled(obj->regionId);

        obj->pnameserverEntry = (FrameQ_NameServerEntry *)virtAddr;

        sharedShmBase = SharedRegion_getSRPtr((Ptr)virtAddr,
                                              obj->regionId);
        nameserverEntry.ctrlInterfaceType =
                                       params->commonCreateParams.ctrlInterfaceType;
        nameserverEntry.instParams.shMemParams.sharedAddr = (Ptr)
                                                          sharedShmBase;
        nameserverEntry.instParams.shMemParams.sharedAddrSize =
                                                 params->sharedAddrSize;
        if (params->sharedAddr != NULL) {
            nameserverEntry.instParams.shMemParams.sharedAddrSize =
                                                     params->sharedAddrSize;
        }
        else {
            nameserverEntry.instParams.shMemParams.sharedAddrSize =
                                                     obj->allocSize;
        }

        if (params->frameQBufMgrName != NULL) {
            strcpy((String)nameserverEntry.instParams.shMemParams.frameQBufMgrName,
                   params->frameQBufMgrName);
        }
        else {
            nameserverEntry.instParams.shMemParams.frameQBufMgrName[0] = '\0';
        }

        if (params->frameQBufMgrSharedAddr != NULL) {
            index = SharedRegion_getId(params->frameQBufMgrSharedAddr);
            nameserverEntry.instParams.shMemParams.frameQBufMgrSharedAddr =
                  (Ptr) SharedRegion_getSRPtr(params->frameQBufMgrSharedAddr,
                                             index);
        }
        else {
            nameserverEntry.instParams.shMemParams.frameQBufMgrSharedAddr =
                                                                           NULL;
        }

        /***********************************************************************
         * Create instance gate based on the parameters if user has not provided
         * gate.
         ***********************************************************************
         */
        if (params->gate != NULL) {
            obj->gate = params->gate;
            nameserverEntry.instParams.shMemParams.instGateMPAddr =
                       (Ptr) ti_sdo_ipc_GateMP_getSharedAddr(obj->gate);
            obj->creGate = FALSE;
        }
        else {
            /* Set flag  indicating user has not provided  GateMP */
            if (FrameQ_ShMem_usedefaultgate == TRUE) {
                obj->creGate = FALSE;
                /* Check if FrameQ module has any default gate set. if not use
                 * GateMP's default gate.
                 */
                if (FrameQ_ShMem_module->defaultInstGate != NULL) {
                    gateHandle = FrameQ_ShMem_module->defaultInstGate;
                }
                else {
                    gateHandle = GateMP_getDefaultRemote();
                    Assert_isTrue ((NULL != gateHandle),
                                    FrameQ_ShMem_A_nullHandle);
                }
            }
            else {
                /* Set the flag indicating that gate is created internally */
                obj->creGate = TRUE;
                /*Create gateMP for this instance using the params provided */
                GateMP_Params_init(&gateParams);
                gateParams.localProtect = (GateMP_LocalProtect)
                                                               params->localProtect;
                gateParams.remoteProtect = (GateMP_RemoteProtect)
                                                              params->remoteProtect;
                gateParams.regionId = obj->regionId;
                obj->gateMPAllocSize = GateMP_sharedMemReq(&gateParams);

                obj->instGateSharedAddr =
                gateParams.sharedAddr = Memory_alloc(
                                                SharedRegion_getHeap(obj->regionId),
                                                obj->gateMPAllocSize,
                                                0,
                                                eb);
                gateHandle = GateMP_create (&gateParams);

                Assert_isTrue ((NULL != gateHandle),
                                FrameQ_ShMem_A_nullHandle);
            }
            /* Store the instance gate control address in  name server so that
             * openers can open it .
             */
            obj->gate = (ti_sdo_ipc_GateMP_Handle)gateHandle;
            nameserverEntry.instParams.shMemParams.
                             instGateMPAddr = (Ptr) ti_sdo_ipc_GateMP_getSharedAddr(obj->gate);

        }

        /* Inserting nameserver entry params at the start of the shared
         * addr to allow open by  address.
         */
        pnameserverEntry = obj->pnameserverEntry;
        memcpy (pnameserverEntry,
                (Ptr)&nameserverEntry,
                sizeof (FrameQ_NameServerEntry));
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_wbInv ((Ptr)pnameserverEntry,
                         sizeof(FrameQ_NameServerEntry),
                         Cache_Type_ALL,
                         TRUE);
        }

        obj->minAlign = Memory_getMaxDefaultTypeAlign();

        if (SharedRegion_getCacheLineSize(obj->regionId) > obj->minAlign) {
            obj->minAlign = SharedRegion_getCacheLineSize(obj->regionId);
        }
        /* Move addrOffset by cache aligned size of
         * FrameQ_NameServerEntry.
         */
        addrOffset += ROUND_UP(sizeof (FrameQ_NameServerEntry), obj->minAlign);

        /**************************************************************
         * Open FrameQBufMgr instance using the provided
         * FrameQBufMgr Name. It Must created before creating FrameQ with
         * this FrameQBufMgr name.
         **************************************************************
         */
        frmQBufMgrOpenParams.commonOpenParams.name = params->frameQBufMgrName;
        frmQBufMgrOpenParams.commonOpenParams.sharedAddr =
                             params->frameQBufMgrSharedAddr;
        frmQBufMgrOpenParams.commonOpenParams.cpuAccessFlags  =
                                        params->cpuAccessFlags;
        tmpStatus = FrameQBufMgr_open (
                                     &bufMgrHandle,
                                     &frmQBufMgrOpenParams);
        if (tmpStatus < 0){
            status = tmpStatus;
            Error_raise(eb,
                        FrameQ_ShMem_E_FrameQBufMgrOpen,
                        NULL,
                        0);
        }
        else {
            /* Successfully opened the FrameQBufMgr instance
             */
            obj->frameQBufMgrHandle = bufMgrHandle;
            obj->frameQBufMgrId = FrameQBufMgr_getId(
                                   obj->frameQBufMgrHandle);

            /* Update the cache flags and CpuAccess flags   in the object
             * By querieng FrameQBufMgr in order to optimize calls to FrameQBufMgr's
             * Invalidate and writeback calls
             */
            FrameQ_Shmem_updateCacheFlags(obj,params->cpuAccessFlags);
            obj->attrs = (FrameQ_ShMem_Attrs *)(virtAddr + addrOffset);

            /* Populate the shared control struct */
            pAttrs =  obj->attrs;

            pAttrs->status       = FrameQ_ShMem_CREATED;
            pAttrs->version      = FrameQ_ShMem_VERSION; /* version info */

            pAttrs->numReaders         = params->numReaders;
            /* Number of queues for each reader */
            pAttrs->numQueues          = params->numQueues;
            pAttrs->primaryQueueId     = 0;
            pAttrs->numActiveReaders   = 0;
            pAttrs->frmQBufMgrId = obj->frameQBufMgrId;
            pAttrs->localProtect = params->localProtect;
            pAttrs->remoteProtect = params->remoteProtect;

            addrOffset +=  ROUND_UP(sizeof (FrameQ_ShMem_Attrs),
                                   obj->minAlign);

            /* Populate the obj with the writer client control info poiner
             * in shared memory
             */
            obj->writer =  (FrameQ_ShMem_WriterClient *)(virtAddr + addrOffset);

            addrOffset += ROUND_UP(sizeof (FrameQ_ShMem_WriterClient),
                          obj->minAlign);
            /* Update the readers information */
            obj->reader = Memory_alloc(NULL,
                                       sizeof(UInt32) * pAttrs->numReaders,
                                       0,
                                       eb);
            if (obj->reader == NULL) {
                status = FrameQ_E_ALLOC_MEMORY;
                Error_raise(eb,
                            FrameQ_ShMem_E_Memory,
                            NULL,
                            0);
            }
            else {
                for (i = 0; i <  pAttrs->numReaders; i++) {
                    obj->reader[i] = (FrameQ_ShMem_ReaderClient *)(virtAddr + addrOffset);

                    addrOffset += ROUND_UP(sizeof (FrameQ_ShMem_ReaderClient),
                                          obj->minAlign);
                }
            }

            if (status >= 0) {
                /* Update the Queue information */
                obj->frmQueue = Memory_alloc(NULL,
                                             (sizeof(UInt32) *
                                              pAttrs->numReaders *
                                              pAttrs->numQueues),
                                              0,
                                              eb);

                if (obj->frmQueue == NULL) {
                    status = FrameQ_E_ALLOC_MEMORY;
                    Error_raise(eb,
                                FrameQ_ShMem_E_Memory,
                                NULL,
                                0);
                }
                else {
                    for (i = 0;
                         i <  (pAttrs->numReaders);
                         i++) {
                        obj->frmQueue[i] = (FrameQ_ShMem_FrameQueue *)
                                                         (virtAddr + addrOffset);

                        addrOffset += ROUND_UP(sizeof (FrameQ_ShMem_FrameQueue),
                                      obj->minAlign);
                    }
                }
            }

            /***********************************************************
             * Create ClientNotifyMgr instance used by this FrameQ
             * instance.Create a gateMP using same protection parameters
             * as specfied in FrameQ create and  pass  it to
             * ClientNotifyMgr_create.
             ***********************************************************
             */

            ClientNotifyMgr_Params_init(&clientMgrParams);
            clientMgrParams.numNotifyEntries = pAttrs->numReaders;
            clientMgrParams.numSubNotifyEntries = pAttrs->numQueues;
            clientMgrParams.sharedAddr = (Ptr)
                                      (  virtAddr + addrOffset);

            clientMgrParams.sharedAddrSize =
                ClientNotifyMgr_sharedMemReq (
                       (ClientNotifyMgr_Params*)&clientMgrParams );
            addrOffset += ROUND_UP(clientMgrParams.sharedAddrSize,
                                  obj->minAlign);

            if (params->gate != NULL) {
                clientMgrParams.gate = (GateMP_Handle) params->gate;
                nameserverEntry.instParams.shMemParams.cliGateMPAddr =
                (Ptr) ti_sdo_ipc_GateMP_getSharedAddr((ti_sdo_ipc_GateMP_Handle)clientMgrParams.gate);

                /* Make it null so that finalization won't delete it.
                 */
                obj->clientNotifyMgrGate = params->gate;
                obj->creCliGate = FALSE;
            }
            else {
                if (FrameQ_ShMem_usedefaultgate == TRUE ) {
                    clientMgrParams.gate = (Ptr) gateHandle;
                    nameserverEntry.instParams.shMemParams.cliGateMPAddr =
                    (Ptr) ti_sdo_ipc_GateMP_getSharedAddr((ti_sdo_ipc_GateMP_Handle)clientMgrParams.gate);
                    /* Make it null so that finalization won't delete it.
                     */
                    obj->clientNotifyMgrGate = NULL;
                    obj->creCliGate = FALSE;
                }
                else {
                    /*Create gateMP for the clientNotifyMgr */
                    obj->creCliGate = TRUE;
                    GateMP_Params_init(&gateParams);

                    /* Since ClientNotifyMgr uses this gate in ISR context
                     * currently local protection should be
                     * GateMP_LocalProtect_INTERRUPT.
                     */

                    gateParams.localProtect = (GateMP_LocalProtect) params->localProtect;
                    gateParams.remoteProtect = (GateMP_RemoteProtect)
                                                          params->remoteProtect;
                    gateParams.sharedAddr  = (Ptr)(virtAddr + addrOffset );
                    gateParams.regionId = obj->regionId;
                    gateMPsharedAddrSize = GateMP_sharedMemReq(&gateParams);

                    addrOffset += ROUND_UP(gateMPsharedAddrSize,
                                           obj->minAlign);
                    gateHandle = GateMP_create (&gateParams);

                    Assert_isTrue ((NULL != gateHandle),
                                    FrameQ_ShMem_A_nullHandle);
                    if (gateHandle == NULL) {
                        status = FrameQBufMgr_E_CREATE_GATEMP;
                        Error_raise(eb,
                                    FrameQ_ShMem_E_GateMPCreate,
                                    NULL,
                                    0);
                    }
                    else {
                        clientMgrParams.gate = (Ptr) gateHandle;
                        nameserverEntry.instParams.shMemParams.
                               cliGateMPAddr = (Ptr)SharedRegion_invalidSRPtr();
                        /* Make it null so that finalization won't delete it.
                         */
                        obj->clientNotifyMgrGate =
                                           (ti_sdo_ipc_GateMP_Handle)gateHandle;
                    }
                }
            }

            if (status >= 0) {
                /* Create the client notifyMgr for this instance */
                /* Set the sub notify entries equivalent to
                 * number queues
                 */
                clientMgrParams.numSubNotifyEntries = params->numQueues;

                /* Pass the same name for ClientNotifyMgr
                */
                clientMgrParams.name =
                                (String)params->commonCreateParams.name;
                /* Event no to be used for this instance */
                clientMgrParams.eventNo =
                                   FrameQ_ShMem_NOTIFY_RESERVED_EVENTNO;
                /* This is actually create call */
                clientMgrParams.openFlag = FALSE;
                if (obj->ctrlStructCacheFlag == TRUE) {
                    clientMgrParams.cacheFlags =
                                       ClientNotifyMgr_CONTROL_CACHEUSE;
                }
                cliMgrHandle = ClientNotifyMgr_create (&clientMgrParams);
                Assert_isTrue ((NULL != cliMgrHandle),
                                FrameQ_ShMem_A_nullHandle);
                if (NULL == cliMgrHandle) {
                    status = FrameQ_E_FAIL_CLIENTN0TIFYMGR_CREATE;
                    Error_raise(eb,
                                FrameQ_ShMem_E_ClientNotifyMgrCreate,
                                NULL,
                                0);
                }
                else {
                    obj->clientNotifyMgrHandle = (Ptr)cliMgrHandle;
                }

                if (status >= 0) {
                    obj->frameList = Memory_alloc (
                                              NULL,
                                              sizeof(UInt32)*
                                              pAttrs->numReaders,
                                              0,
                                              eb);
                    if (obj->frameList  == NULL) {
                        status = FrameQ_E_MEMORY;
                        Error_raise(eb,
                                    FrameQ_ShMem_E_Memory,
                                    NULL,
                                    0);
                    }
                    else {
                        for (k = 0;
                             k < pAttrs->numReaders;
                             k++) {
                            obj->frameList[k] = Memory_alloc (
                                               NULL,
                                               sizeof(UInt32)*
                                               pAttrs->numQueues,
                                               0,
                                               eb);
                            if (obj->frameList[k]  == NULL) {
                                status = FrameQ_E_MEMORY;
                                Error_raise(eb,
                                            FrameQ_ShMem_E_Memory,
                                            NULL,
                                            0);
                            }
                        }
                    }

                    if  (status >= 0) {
                        /* Prepare  frame queues.
                         * Pass shared Addr to all the ListMPs
                         * that FrameQ is creating.
                         */
                        ListMP_Params_init (&listMPParams);
                        /* Set the region id in order to calculate the
                         * shared mem required.
                         */
                        listMPParams.regionId = SharedRegion_getId((Ptr)virtAddr);
                        listMPParams.gate = (GateMP_Handle)obj->gate;
                        listMPsharedAddrSize =
                                    ListMP_sharedMemReq (&listMPParams);
                        /*
                         * Each frmQueue entry in attrs maintains info
                         * about queues that belongs  one reader.
                         */
                        for (i = 0;
                             i < pAttrs->numReaders;
                             i++) {
                            obj->frmQueue[i]->isAllocated = FALSE;
                            obj->frmQueue[i]->readerIndex = (UInt32)-1;
                            for (k = 0;
                                 k < pAttrs->numQueues;
                                 k++) {
                                /* Initialize the numframes in each reader
                                 * queue to zero
                                 */
                                obj->frmQueue[i]->numFrames[k] = 0;
                                listMPParams.sharedAddr = (UInt32 *)
                                          ( (UInt32) virtAddr
                                           + addrOffset);
                                obj->frameList[i][k] = (ti_sdo_ipc_ListMP_Object*) ListMP_create (
                                                          &listMPParams);
                                addrOffset += listMPsharedAddrSize;
                            }
                            if (obj->ctrlStructCacheFlag == TRUE) {
                                Cache_wbInv ((Ptr)obj->frmQueue[i],
                                             sizeof(FrameQ_ShMem_FrameQueue),
                                             Cache_Type_ALL,
                                             TRUE);

                            }
                        }
                    }
                }

                if (status >= 0) {
                    /* Initialize writer client structure */
                    writer =
                        (FrameQ_ShMem_WriterClient *)obj->writer;
                    writer->isValid     = FALSE ;
                    writer->procId      = (UInt32)-1;
                    writer->cacheFlags  = 0;
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_wbInv ((Ptr)writer,
                                     sizeof(FrameQ_ShMem_WriterClient),
                                     Cache_Type_ALL,
                                     TRUE);

                    }
                    /* Initialize reader clients control structures */
                    for (i = 0; i < pAttrs->numReaders; i++) {
                        obj->reader[i]->isValid      = FALSE;
                        obj->reader[i]->procId       = (UInt32)-1;
                        obj->reader[i]->frameQIndex  = (UInt32)-1;
                        obj->reader[i]->cacheFlags   = 0;
                        obj->reader[i]->isRegistered = FALSE;
                        obj->reader[i]->notifyId     = (UInt32)-1;
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_wbInv ((Ptr)obj->reader[i],
                                         sizeof(FrameQ_ShMem_WriterClient),
                                         Cache_Type_ALL,
                                         TRUE);

                        }
                    }

                    /* Write back Attrs structure */
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_wbInv ((Ptr)pAttrs,
                                     sizeof(FrameQ_ShMem_Attrs),
                                     Cache_Type_ALL,
                                     TRUE);
                    }

                    if (params->sharedAddr) {
                        obj->objType   = FrameQ_ShMem_DYNAMIC_CREATE;
                    }
                    else {
                        /* set the object type indicating it is created
                         * using Shared Region id internally.
                         */
                        obj->objType  =
                                 FrameQ_ShMem_DYNAMIC_CREATE_USEDREGION;
                    }
                    obj->ctrlInterfaceType =
                                 params->commonCreateParams.ctrlInterfaceType;
                    obj->virtClientHandle = NULL;
                    /**********************************************************
                     * Add nameserverEntry in to the name server instance
                     **********************************************************
                     */
                    if ((FrameQ_ShMem_module->nameServer != NULL) &&
                        (strlen((String)params->commonCreateParams.name) > 0)) {
                        /* FrameQ_ShMem_module->nameServer is the same nameserver
                         * instance created by the front end FrameQ module.
                         */
                        obj->nsKey = NameServer_add (
                                       (NameServer_Handle)FrameQ_ShMem_module->nameServer,
                                       (String)params->commonCreateParams.name,
                                       &nameserverEntry,
                                       sizeof(FrameQ_NameServerEntry));
                        if (obj->nsKey == NULL) {
                            status = FrameQ_E_NAMESERVERADD;
                            Assert_isTrue ((NULL != obj->nsKey),
                                           FrameQ_ShMem_A_nullPointer);

                        }
                    }
                }
            }
        }
    }

    if (   (strlen((String)params->commonCreateParams.name) > 0)
        && (status >= 0)) {
        obj->name = Memory_alloc (NULL,
                                  FrameQ_ShMem_MAX_NAMELEN,
                                  0,
                                  eb);
        if (obj->name == NULL) {
            status = FrameQ_E_MEMORY;
            Error_raise(eb,
                        FrameQ_ShMem_E_Memory,
                        NULL,
                        0);
        }
        else {
            if (  strlen((String)params->commonCreateParams.name)
                < FrameQ_ShMem_MAX_NAMELEN) {
                strcpy(obj->name, (String)params->commonCreateParams.name);
            }
        }

    }

    return (status);
}

/*
 *  ======== _FrameQ_ShMem_open ========
 *  Internal API to open the instance.
 */
static inline
Int32
_FrameQ_ShMem_open (FrameQ_ShMem_Object       *obj,
                    const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params *params,
                    Error_Block               *eb)
{
    Int32                   status      = FrameQ_S_SUCCESS;
    UInt32                  virtAddr    = 0;
    UInt32                  addrOffset   = 0;
    Int32                   tmpStatus;
    volatile FrameQ_ShMem_Attrs  *pAttrs;
    ListMP_Params           listMPParams;
    UInt32                  i;
    UInt32                  key;
    GateMP_Params           gateParams;
    GateMP_Handle           gateHandle;
    ClientNotifyMgr_Params  clientMgrParams;
    ClientNotifyMgr_Handle  cliMgrHandle;
    FrameQ_NameServerEntry  nameserverEntry;
    FrameQ_NameServerEntry  *pnameserverEntry;
    UInt32                  k;
    FrameQBufMgr_OpenParams frmQBufMgrOpenParams;
    FrameQBufMgr_Handle     bufMgrHandle;
    UInt32                  listMPsharedAddrSize;
    Ptr                     listMPsharedAddr;
    Ptr                     localAddr;
    Int32                   index;
    UInt32                  len;

    Assert_isTrue ((NULL != params), FrameQ_ShMem_A_invalidParams);

    /* Get the shared control addr , if NameServer is configured  and
     * name is provided in the params.
     */
    if (   (strlen((String)params->commonCreateParams.name) > 0)
        && (FrameQ_ShMem_module->nameServer != NULL)) {

        len = sizeof (FrameQ_NameServerEntry);
        status = NameServer_get ((NameServer_Handle)FrameQ_ShMem_module->nameServer,
                                 (String)(params->commonCreateParams.name),
                                 &nameserverEntry,
                                 &len,
                                 NULL/* TODO: check need to pass multiproc
                                        info*/);
        if (status == NameServer_E_NOTFOUND) {
            /* Name not found. */
            /* Failed to get the information from the name server.
             * It means it is not created.
             */
            if (params->sharedAddr == NULL) {
                status = FrameQ_E_NOTFOUND;
                Error_raise(eb,
                            FrameQ_ShMem_E_FrameQNotFound,
                            NULL,
                            0);
            }
        }
        else if (status == NameServer_E_FAIL) {
            /* Error happened in NameServer. Pass the error up. */
            status = FrameQ_E_FAIL;
            Error_raise(eb,
                        FrameQ_ShMem_E_Fail,
                        NULL,
                        0);
        }
        else {
            virtAddr = (UInt32) SharedRegion_getPtr (
                                      (SharedRegion_SRPtr)
                            nameserverEntry.instParams.shMemParams.sharedAddr);
            pnameserverEntry = &nameserverEntry;
            /* Assert that sharedAddr is cache aligned */
            index = SharedRegion_getId((Ptr)virtAddr);

            obj->regionId = index;
            /* Assert that the buffer is in a valid shared region */
            Assert_isTrue(index != SharedRegion_INVALIDREGIONID,
                          Ipc_A_addrNotInSharedRegion);

            Assert_isTrue(((UInt32)virtAddr %
                          SharedRegion_getCacheLineSize(index) == 0),
                          Ipc_A_addrNotCacheAligned);

            if (SharedRegion_isCacheEnabled(index)) {
                obj->ctrlStructCacheFlag = TRUE;
                Cache_inv((Ptr)virtAddr,
                          ROUND_UP (
                                  sizeof (FrameQ_NameServerEntry),
                                  obj->minAlign),
                          Cache_Type_ALL,
                          TRUE);
            }

        }
    }

    if (virtAddr == 0 ) {
        if (params->sharedAddr != NULL) {
            /* Unable to get information from name server. But user has
             * provided  virtual shared address. So proceeding further.
             */
            virtAddr    = (UInt32) params->sharedAddr;
            pnameserverEntry = (FrameQ_NameServerEntry*)virtAddr;
            /* Assert that sharedAddr is cache aligned */
            index = SharedRegion_getId(pnameserverEntry);

            /* Assert that the buffer is in a valid shared region */
            Assert_isTrue(index != SharedRegion_INVALIDREGIONID,
                          Ipc_A_addrNotInSharedRegion);

            Assert_isTrue(((UInt32)pnameserverEntry %
                          SharedRegion_getCacheLineSize(index) == 0),
                          Ipc_A_addrNotCacheAligned);
            /* Invalidate FrameQ enry info located at the start of
             * the sharedAddr.
             */
            if (SharedRegion_isCacheEnabled(index)) {
                Cache_inv((Ptr)pnameserverEntry,
                          ROUND_UP (
                                  sizeof (FrameQ_NameServerEntry),
                                  obj->minAlign),
                          Cache_Type_ALL,
                          TRUE);
                obj->ctrlStructCacheFlag = TRUE;
            }
            status  = FrameQ_S_SUCCESS;
        }
        else {
            /*
             * params->sharedAddr is null and also virtaddr got from the
             * name server is  also not valid. So rejecting the  open
             * call.
             */
            status = FrameQ_E_FAIL;
            Error_raise(eb,
                        FrameQ_ShMem_E_FrameQNotCreated,
                        NULL,
                        0);
        }
    }
    /* In shared memory  lay out  for FrameQ instance is
     * FrameQ_NameServerEntry - Cache aligned size.
     * FrameQ conrol struct
     * ClientNoitifyMgr gate
     * ClientNotifyMgr control struct
     * listMp objects (numreaders * numqueues)
     */
    if (status >= 0) {
        Assert_isTrue (   (nameserverEntry.instParams.shMemParams.instGateMPAddr
                != (Ptr)SharedRegion_invalidSRPtr()),
                       FrameQ_ShMem_A_invalidParams);
        if (   nameserverEntry.instParams.shMemParams.instGateMPAddr
            != (Ptr)SharedRegion_invalidSRPtr())
        /* Get the local address of the SRPtr */
        localAddr = SharedRegion_getPtr((SharedRegion_SRPtr)
                     nameserverEntry.instParams.shMemParams.instGateMPAddr);
        Assert_isTrue ((localAddr != NULL),
                        FrameQ_ShMem_A_nullPointer);

        tmpStatus = GateMP_openByAddr(localAddr, &gateHandle);
        if (tmpStatus != GateMP_S_SUCCESS) {
            obj->creGate = FALSE;
            status = FrameQ_E_OPEN_GATEMP;
            Error_raise(eb,
                        FrameQ_ShMem_E_GateMPOpen,
                        NULL,
                        0);
        }
        else {
            obj->gate    = (ti_sdo_ipc_GateMP_Handle)gateHandle;
            obj->creGate = TRUE;
        }
    }

    if (status >= 0) {

        obj->minAlign = Memory_getMaxDefaultTypeAlign();

        obj->regionId = SharedRegion_getId((Ptr)virtAddr);

        if (SharedRegion_getCacheLineSize(obj->regionId) > obj->minAlign) {
            obj->minAlign = SharedRegion_getCacheLineSize(obj->regionId);
        }

        /* Move offset by size of nameserverEntry cache aligned size.
         * No need to check interface type as it is checked by front end
         * module(FrameQ).
         */
        obj->pnameserverEntry = (FrameQ_NameServerEntry*)(virtAddr);
        addrOffset += ROUND_UP(sizeof (FrameQ_NameServerEntry), obj->minAlign);

        obj->attrs = (FrameQ_ShMem_Attrs *)(virtAddr + addrOffset);

        addrOffset +=  ROUND_UP(sizeof (FrameQ_ShMem_Attrs),
                               obj->minAlign);

        pAttrs       = obj->attrs;

        /***************************************************
         * Open the FrameQBufMgr instance using the provided
         * FrameQBufMgr Name.
         ***************************************************
         */
        frmQBufMgrOpenParams.commonOpenParams.name = (String)
            pnameserverEntry->instParams.shMemParams.frameQBufMgrName;
        frmQBufMgrOpenParams.commonOpenParams.sharedAddr =
                pnameserverEntry->instParams.shMemParams.
                                     frameQBufMgrSharedAddr;
        frmQBufMgrOpenParams.commonOpenParams.cpuAccessFlags  =
                                        params->cpuAccessFlags;
        tmpStatus = FrameQBufMgr_open (&bufMgrHandle,
                                       &frmQBufMgrOpenParams);
        if (tmpStatus < 0) {
            status = FrameQ_E_FRAMEQBUFMGROPEN;
            Error_raise(eb,
                        FrameQ_ShMem_E_FrameQBufMgrOpen,
                        NULL,
                        0);
        }
        else {
            /* Successfully opened the FrameQBufMgr instance
             */
            obj->frameQBufMgrHandle = bufMgrHandle;
            obj->frameQBufMgrId = FrameQBufMgr_getId(
                                   obj->frameQBufMgrHandle);
            /* Update the cache flags and CpuAccess flags   in the object
             * By querieng FrameQBufMgr in order to optimize calls to
             * FrameQBufMgr's invalidate and writeback calls.
             */
            FrameQ_Shmem_updateCacheFlags(obj,params->cpuAccessFlags);

            key = GateMP_enter ((GateMP_Handle)obj->gate);

            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv((Ptr)pAttrs,
                          sizeof(FrameQ_ShMem_Attrs),
                          Cache_Type_ALL,
                          TRUE);
            }

            if (pAttrs->status != FrameQ_ShMem_CREATED) {
                status = FrameQ_E_NOTFOUND;
                Error_raise(eb,
                            FrameQ_ShMem_E_FrameQNotCreated,
                            NULL,
                            0);
            }
            else {

                /* Update the writer information */
                obj->writer =  (FrameQ_ShMem_WriterClient *)(virtAddr + addrOffset);

                addrOffset += ROUND_UP(sizeof (FrameQ_ShMem_WriterClient),
                              obj->minAlign);

                /* Update the readers information */
                obj->reader = Memory_alloc(NULL,
                                           sizeof(UInt32) * pAttrs->numReaders,
                                           0,
                                           eb);
                if (obj->reader == NULL) {
                    status = FrameQ_E_ALLOC_MEMORY;
                    Error_raise(eb,
                                FrameQ_ShMem_E_Memory,
                                NULL,
                                0);
                }
                else {
                    for (i = 0; i <  pAttrs->numReaders; i++) {
                        obj->reader[i] = (FrameQ_ShMem_ReaderClient *)(virtAddr + addrOffset);

                        addrOffset += ROUND_UP(sizeof (FrameQ_ShMem_ReaderClient),
                                      obj->minAlign);
                    }
                }

                if (status >= 0) {
                    /* Update the Queue information */
                    obj->frmQueue = Memory_alloc(NULL,
                                                 (sizeof(UInt32) *
                                                  pAttrs->numReaders *
                                                  pAttrs->numQueues),
                                                  0,
                                                  eb);

                    if (obj->frmQueue == NULL) {
                        status = FrameQ_E_ALLOC_MEMORY;
                        Error_raise(eb,
                                    FrameQ_ShMem_E_Memory,
                                    NULL,
                                    0);
                    }
                    else {
                        for (i = 0;
                             i <  (pAttrs->numReaders);
                             i++) {
                            obj->frmQueue[i] = (FrameQ_ShMem_FrameQueue *)
                                                        (virtAddr + addrOffset);

                            addrOffset += ROUND_UP(
                                               sizeof (FrameQ_ShMem_FrameQueue),
                                               obj->minAlign);
                        }
                    }
                }

                /* Get the default params from the client notify manager */
                ClientNotifyMgr_Params_init(&clientMgrParams);
                clientMgrParams.numNotifyEntries = pAttrs->numReaders;
                clientMgrParams.numSubNotifyEntries = pAttrs->numQueues;
                clientMgrParams.sharedAddr = (Ptr) (virtAddr + addrOffset);
                clientMgrParams.sharedAddrSize =
                        ClientNotifyMgr_sharedMemReq (
                           (ClientNotifyMgr_Params*)&clientMgrParams );

                addrOffset += ROUND_UP(clientMgrParams.sharedAddrSize,
                                      obj->minAlign);
                /* cliGateMPAddr if it valid means user has passed the
                 * gate to create call or  default gate is used in create call.
                 */
                if (   nameserverEntry.instParams.shMemParams.cliGateMPAddr
                    != (Ptr)SharedRegion_invalidSRPtr()) {
                    /* Get the local address of the SRPtr */
                    localAddr = SharedRegion_getPtr((SharedRegion_SRPtr)
                          nameserverEntry.instParams.shMemParams.cliGateMPAddr);
                    Assert_isTrue ((localAddr != NULL),
                            FrameQ_ShMem_A_nullPointer);
                    tmpStatus = GateMP_openByAddr(localAddr, &gateHandle);
                    if (tmpStatus != GateMP_S_SUCCESS) {
                        status = FrameQ_E_OPEN_GATEMP;
                        obj->creCliGate = FALSE;
                        /* Make it null so that finalization won't delete it.
                         */
                        obj->clientNotifyMgrGate = NULL;
                        }
                    else {
                        clientMgrParams.gate = (Ptr)gateHandle;
                        obj->clientNotifyMgrGate  = (ti_sdo_ipc_GateMP_Handle)
                                                        gateHandle;
                        obj->creCliGate = TRUE;
                    }
                }
                else {
                    /* Open the gate created for the ClientNotifyMgr
                     * instance.
                     */
                    obj->creCliGate = TRUE;
                    GateMP_Params_init(&gateParams);

                    gateParams.localProtect  =
                                  (GateMP_LocalProtect)pAttrs->localProtect;
                    gateParams.remoteProtect =
                                (GateMP_RemoteProtect)pAttrs->remoteProtect;

                    gateParams.sharedAddr  = (Ptr)(virtAddr + addrOffset );
                    gateParams.regionId = obj->regionId;
                    addrOffset += ROUND_UP(GateMP_sharedMemReq(&gateParams),
                                           obj->minAlign);
                    tmpStatus = GateMP_openByAddr(gateParams.sharedAddr,
                                              &gateHandle);
                    if (tmpStatus != GateMP_S_SUCCESS) {
                        status = FrameQ_E_OPEN_GATEMP;
                        Error_raise(eb,
                                    FrameQ_ShMem_E_GateMPOpen,
                                    NULL,
                                    0);
                    }
                    else {
                        clientMgrParams.gate = (Ptr)gateHandle;
                        obj->clientNotifyMgrGate  = (ti_sdo_ipc_GateMP_Handle)
                                                                 gateHandle;
                    }
                }
                if (status >= 0) {
                    if ( obj->ctrlStructCacheFlag == TRUE) {
                        clientMgrParams.cacheFlags =
                                           ClientNotifyMgr_CONTROL_CACHEUSE;
                    }

                    /*Open the client notifyMgr instace*/
                    clientMgrParams.openFlag = TRUE;

                    cliMgrHandle = ClientNotifyMgr_create (&clientMgrParams);
                    if (cliMgrHandle == NULL) {
                        obj->clientNotifyMgrHandle = NULL;
                        status = FrameQ_E_FAIL_CLIENTN0TIFYMGR_OPEN;
                        Error_raise(eb,
                                    FrameQ_ShMem_E_ClientNotifyMgrOpen,
                                    NULL,
                                    0);
                    }
                    else {
                        obj->clientNotifyMgrHandle = (Ptr)cliMgrHandle;
                        obj->objType       = FrameQ_ShMem_DYNAMIC_OPEN;
                        obj->ctrlInterfaceType =
                                     params->commonCreateParams.ctrlInterfaceType;
                        obj->frameList = Memory_alloc (
                                                     NULL,
                                                     sizeof(UInt32)*
                                                     pAttrs->numReaders,
                                                     0,
                                                     eb);
                        if (obj->frameList  == NULL) {
                            status = FrameQ_E_MEMORY;
                            Error_raise(eb,
                                        FrameQ_ShMem_E_Memory,
                                        NULL,
                                        0);
                        }
                        else {
                            for (k = 0;
                                 k < pAttrs->numReaders;
                                 k++) {
                                obj->frameList[k] = Memory_alloc (
                                               NULL,
                                               sizeof(UInt32)*
                                               pAttrs->numQueues,
                                               0,
                                               eb);
                                if (obj->frameList[k]  == NULL) {
                                    status = FrameQ_E_MEMORY;
                                    Error_raise(eb,
                                                FrameQ_ShMem_E_Memory,
                                                NULL,
                                                0);
                                }
                            }
                        }

                        if  (status >= 0) {
                            ListMP_Params_init(&listMPParams);
                            /* Set the region id in order to calculate the
                             * shared mem required.
                             */
                            listMPParams.regionId = SharedRegion_getId((Ptr)virtAddr);
                            listMPParams.gate = (GateMP_Handle)obj->gate;
                            listMPsharedAddrSize =
                            ListMP_sharedMemReq(&listMPParams);

                            /*
                             * Each frmQueue entry in attrs maintains info
                             * about queues that belongs  one reader.
                             * Open all the listMPs created for this frameQ.
                             */
                            for (i = 0;
                                 i < pAttrs->numReaders;
                                 i++) {
                                for (k = 0;
                                     k < pAttrs->numQueues;
                                     k++) {
                                    /* Open the ListMP by using sharedAddr
                                     * as code will become complex if we
                                     * start using names for all the listMPs
                                     */
                                    listMPsharedAddr = (Ptr)
                                                    ( (UInt32)virtAddr + addrOffset);
                                    tmpStatus  =  ListMP_openByAddr (
                                                    listMPsharedAddr,
                                                    (ListMP_Handle*)&(obj->frameList[i][k]));
                                    if (tmpStatus < 0){
                                        status = FrameQ_E_LISTMPOPEN;
                                        break;
                                    }
                                    addrOffset += ROUND_UP(listMPsharedAddrSize,
                                                          obj->minAlign);
                                }
                            }
                        }
                    }
                }
            }

            GateMP_leave ((GateMP_Handle)obj->gate, key);
        }
    }

    if (   (strlen((String)params->commonCreateParams.name) > 0)
        && (status >= 0)) {
        obj->name = Memory_alloc (NULL,
                                  FrameQ_ShMem_MAX_NAMELEN,
                                  0,
                                  eb);
        if (obj->name == NULL) {
            status = FrameQ_E_MEMORY;
            Error_raise(eb,
                        FrameQ_ShMem_E_Memory,
                        NULL,
                        0);
        }
        else {
            if (  strlen((String)params->commonCreateParams.name)
                < FrameQ_ShMem_MAX_NAMELEN) {
                strcpy(obj->name, (String)params->commonCreateParams.name);
            }
        }

    }

    return (status);
}

/*
 * ======== FrameQ_ShMem_readerPutv ========
 * Function to insert multiple frames  back in to the queue of a
 * FrameQ reader client. Internal API.
 */
static inline Int32
FrameQ_ShMem_readerPutv (FrameQ_ShMem_Object    *obj,
                         FrameQ_ShMem_Frame     framePtr[],
                         UInt32                  filledQueueId[],
                         UInt8                  numFrames)
{
    Int32                  status = FrameQ_S_SUCCESS;
    UInt32                 i            = 0;
    FrameQ_ShMem_Attrs         *pAttrs       = NULL;
    FrameQ_ShMem_ReaderClient  *reader;
    FrameQ_ShMem_FrameQueue    *queue;
    UInt32                 key;
    FrameQBufMgr_Handle    bufMgrHandle;
    Ptr                    portablePtr;
    UInt8                  filledQueueNo;
    UInt32                 k;

    pAttrs = (FrameQ_ShMem_Attrs *) obj->attrs;

    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)pAttrs,
                   sizeof (FrameQ_ShMem_Attrs),
                   Cache_Type_ALL,
                   TRUE);
    }

    reader = obj->virtClientHandle;
    queue = (FrameQ_ShMem_FrameQueue *)obj->frmQueue[reader->frameQIndex];
    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv (queue,
                   sizeof (FrameQ_ShMem_FrameQueue),
                   Cache_Type_ALL,
                   TRUE);
    }
    if ((queue->isAllocated == TRUE)) {
        for (i = 0; i < numFrames; i++) {
            /* If client happens to be  reader, then just put the frame at
             * the head of the corresponding reader's queue
             */
            /* Convert all address to sharead region address in
             * frame header.i.e frame buffer pointers
             */
            bufMgrHandle = obj->frameQBufMgrHandle;
            Assert_isTrue ((NULL != bufMgrHandle),
                           FrameQ_ShMem_A_invalidFrameQBufMgrId);

            for (k = 0; k < framePtr[i]->numFrameBuffers; k++) {
                if (   (obj->frmBufCacheFlag[k] == TRUE)
                    && (obj->bufCpuAccessFlag[k] == TRUE)) {
                status = FrameQBufMgr_writeBackFrameBuf(
                                bufMgrHandle,
                               (Ptr)
       (framePtr[i]->frameBufInfo[k].bufPtr + framePtr[i]->frameBufInfo[k].startOffset),
                               framePtr[i]->frameBufInfo[k].validSize,
                               k);
                }
                /* Address translation to portable pointers */
                status =  FrameQBufMgr_translateAddr (
                                    bufMgrHandle,
                                    &portablePtr,
                                    FrameQBufMgr_AddrType_Portable,
                                    (Ptr)framePtr[i]->frameBufInfo[k].bufPtr,
                                    FrameQBufMgr_AddrType_Virtual,
                                    FrameQBufMgr_BUF_FRAMEBUF);

                Assert_isTrue ( (NULL != portablePtr),
                                 FrameQ_ShMem_A_nullPointer);
                framePtr[i]->frameBufInfo[k].bufPtr = (UInt32)portablePtr;
            }

            if (obj->frmHdrBufCacheFlag == TRUE) {
                status = FrameQBufMgr_writeBackHeaderBuf(bufMgrHandle,
                                                         framePtr[i]);
            }

            filledQueueNo = filledQueueId[i];
            /* Put in to filled queue of the reader */
            ListMP_putHead ((ListMP_Handle)obj->frameList[reader->frameQIndex][filledQueueNo] ,
                           (ListMP_Elem *)framePtr[i]);
            queue->numFrames[filledQueueNo]++;
        }


        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_wbInv (queue,
                         sizeof (FrameQ_ShMem_FrameQueue),
                         Cache_Type_ALL,
                         TRUE);
        }

    }
    else {
        status = FrameQ_E_FAIL;
    }

    /* Release lock*/
    GateMP_leave ((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 * ======== FrameQ_ShMem_writerPutv ========
 * Internal API. Function to insert multiple frames in to the queues of
 * reader client. If multiple readers are active at the same time, it dups and
 * inserts the frames. If  it able to dup successfully, then only it inserts
 * frames other wise it returns failure.
 *
 */
static inline Int32
FrameQ_ShMem_writerPutv (FrameQ_ShMem_Object    *obj,
                         FrameQ_ShMem_Frame     framePtr[],
                         UInt32                  filledQueueId[],
                         UInt8                  numFrames)
{
    Int32                       status       = FrameQ_S_SUCCESS;
    UInt32                      i            = 0;
    FrameQ_ShMem_Attrs         *pAttrs       = NULL;
    FrameQ_ShMem_Frame          pframe[FrameQ_MAX_FRAMESINVAPI][FrameQ_MAX_INST_READERS]  = { NULL};
    FrameQ_ShMem_Frame          temp         = NULL;
    Int32                       tmpStatus    = FrameQ_S_SUCCESS;
    FrameQ_ShMem_ReaderClient  *reader;
    FrameQ_ShMem_FrameQueue    *queue;
    FrameQBufMgr_Handle         bufMgrHandle;
    Ptr                         portablePtr;
    UInt8                       filledQueueNo;
    UInt32                      k;
    UInt32                      rdrCnt;
    UInt32                      counter;
    UInt32                      key;
    Error_Block                 eb;

    Assert_isTrue ((FrameQ_MAX_FRAMESINVAPI >= numFrames), FrameQ_ShMem_A_invalidParams);

    /* Check if Frames passed are allocated from the FrameQBufMgr that this
     * writer supports.
     */
    for (i = 0; i < numFrames; i ++) {
        if  (framePtr[i]->frmAllocaterId != obj->frameQBufMgrId) {
            /* Frame can not be inserted as it ia allocated from different FramQ
             *BufMgr
             */
            status =  FrameQ_E_INVALID_FRAMEQBUFMGRID;
        }
    }

    if (status >= 0) {
        Error_init(&eb);

        pAttrs = (FrameQ_ShMem_Attrs *) obj->attrs;

        key = GateMP_enter ((GateMP_Handle)obj->gate);

        /* Invalidate to get the updated attrs  */
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)pAttrs,
                       sizeof (FrameQ_ShMem_Attrs),
                       Cache_Type_ALL,
                       TRUE);
        }

        bufMgrHandle = obj->frameQBufMgrHandle;
        Assert_isTrue ((NULL != bufMgrHandle),
                       FrameQ_ShMem_A_invalidFrameQBufMgrId);

        if (pAttrs->numActiveReaders > 1) {
            /*More than one reader exist. Need to duplicate frames to insert
             *frames in to queues other than primary queue.
             */
             /* Exclude primary reader(first reader) in dups as frames passed by APPs
              * will be put in to the primary queue.
              */
             /* Allocate frames  by duplicating. Need
              * to allocate as many as there are number of active readers
              * except primary queue reader.
              */

             tmpStatus = FrameQBufMgr_dupv (obj->frameQBufMgrHandle,
                                            (FrameQBufMgr_Frame *)framePtr,
                                            (FrameQBufMgr_Frame **)pframe,
                                            pAttrs->numActiveReaders -1,
                                            numFrames);
             if (tmpStatus < 0) {
                 status = tmpStatus;
             }
        }


        if (status >= 0) {
            /* Write back the original frames */
            for (i = 0; i < numFrames; i++) {
                /* Translate the frame buffer address to portable addreess */
                for (k = 0; k < framePtr[i]->numFrameBuffers; k++) {
                    if (   (obj->frmBufCacheFlag[k] == TRUE)
                        && (obj->bufCpuAccessFlag[k] == TRUE)) {
                        status = FrameQBufMgr_writeBackFrameBuf(
                                   bufMgrHandle,
                                   (Ptr)
           (framePtr[i]->frameBufInfo[k].bufPtr + framePtr[i]->frameBufInfo[k].startOffset),
                                   framePtr[i]->frameBufInfo[k].validSize,
                                   k);
                    }
                    /* Address translation to portable pointers */
                    status =  FrameQBufMgr_translateAddr (
                                        bufMgrHandle,
                                        &portablePtr,
                                        FrameQBufMgr_AddrType_Portable,
                                        (Ptr)framePtr[i]->frameBufInfo[k].bufPtr,
                                        (FrameQBufMgr_AddrType)FrameQBufMgr_AddrType_Virtual,
                                        FrameQBufMgr_BUF_FRAMEBUF);

                    Assert_isTrue ( (NULL != portablePtr),
                                     FrameQ_ShMem_A_nullPointer);
                    framePtr[i]->frameBufInfo[k].bufPtr = (UInt32)portablePtr;
                }

                if (obj->frmHdrBufCacheFlag == TRUE) {
                    status = FrameQBufMgr_writeBackHeaderBuf(bufMgrHandle,
                                                             framePtr[i]);
                }

                /*Copy the contents of original frame header contents to other
                 * duped frame headers.
                 */
                if (pAttrs->numActiveReaders > 1) {
                    for (rdrCnt = 0;
                         rdrCnt < pAttrs->numActiveReaders -1;
                         rdrCnt++) {
                        memcpy (pframe[i][rdrCnt],
                                framePtr[i],
                                framePtr[i]->headerSize);
                    }
                }

            }
        }

        /* Now insert frames in to readers queues */
        if (status >= 0) {
            for (i = 0; i < numFrames; i++) {
                /* Retrive the filledQueue No */
                filledQueueNo = filledQueueId[i];
                counter = 0;

                 for (  k = 0; k < pAttrs->numReaders; k++) {
                    if( k != pAttrs->primaryQueueId) {
                        queue = (FrameQ_ShMem_FrameQueue *)
                                                     obj->frmQueue[k];
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_inv(queue,
                                      sizeof (FrameQ_ShMem_FrameQueue),
                                      Cache_Type_ALL,
                                      TRUE);
                        }

                        temp = (FrameQ_ShMem_Frame)pframe[i][k];
                    }
                    else {
                        /* Insert the original frame in to primary queue
                         * irrespective of numActive readers
                         */
                        queue = (FrameQ_ShMem_FrameQueue *)
                                obj->frmQueue[pAttrs->primaryQueueId];
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_inv(queue,
                                      sizeof (FrameQ_ShMem_FrameQueue),
                                      Cache_Type_ALL,
                                      TRUE);
                        }
                        temp = framePtr[i];

                    }
                    if (   (queue->isAllocated == TRUE)
                        || (k== pAttrs->primaryQueueId)) {
                        /* Allow insertion in to primary queue evene if it
                         * is not allocated to any reader.
                         */
                        counter++;
                        /* Increment the frame count of filled queue 0 of
                         * reader.
                         */
                        queue->numFrames[filledQueueNo]++;
                        /*TODO: what about frame header pointer.ListMP is
                         * assuming shared region for list elements.
                         * if header buffers are  not from shared region it
                         * will give problem.
                         */
                        ListMP_putTail ((ListMP_Handle)obj->frameList[k][filledQueueNo] ,
                                        (ListMP_Elem *)temp);
                        if (obj->ctrlStructCacheFlag == TRUE) {
                            Cache_wbInv(queue,
                                        sizeof (FrameQ_ShMem_FrameQueue),
                                        Cache_Type_ALL,
                                        TRUE);
                        }
                        /* Check if queue is allocated  to reader then only do
                         * invalidation of reader to send notifications. There
                         * may be a case where reader is not opened yet hence
                         * readerIndex may be -1.
                         */
                        if (queue->isAllocated == TRUE) {
                            /* if queue is allocated queue->readerIndex is a valid
                             * one
                             */
                            /* Send Notification to the Reader if it
                             * is registered for notification
                             */
                            reader = (FrameQ_ShMem_ReaderClient*)
                                    obj->reader[queue->readerIndex];
                            if (obj->ctrlStructCacheFlag == TRUE) {
                                 Cache_inv(reader,
                                           sizeof (FrameQ_ShMem_ReaderClient),
                                           Cache_Type_ALL,
                                           TRUE);
                            }
                            if (reader->isRegistered == TRUE) {
                                /* Release lock */
                                GateMP_leave ((GateMP_Handle)obj->gate, key);
                                status =
                                    ClientNotifyMgr_sendNotification (
                                                     (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                     reader->notifyId,
                                                     filledQueueNo,
                                                     queue->numFrames[filledQueueNo],
                                                     (UInt32)-1);

                                key = GateMP_enter ((GateMP_Handle)obj->gate);

                            }
                        }
                    }

                    if (counter == pAttrs->numActiveReaders) {
                        break;
                    }
                 }
             }
        }

        /* Release lock */
        GateMP_leave ((GateMP_Handle)obj->gate, key);
    }
    return (status);
}

UInt32
ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_sharedMemReq__F (const ti_syslink_ipc_rtos_shMem_FrameQ_ShMem_Params *params)
{
    FrameQ_ShMem_Params shparams;
    /*TBD: Copy shared params from RTSC params */

    return FrameQ_ShMem_sharedMemReq(&shparams);
}
